# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest  # noqa: F401
import torch

import theseus as th


class ResidualCostFunction(th.CostFunction):
    def __init__(
        self,
        optim_vars,
        cost_weight,
        name=None,
        true_coeffs=None,
        point=None,
        multivar=False,
        noise_mag=0,
        target=None,
    ):
        super().__init__(cost_weight, name=name)
        len_vars = len(optim_vars) if multivar else optim_vars[0].dof()
        self._optim_vars = optim_vars

        if isinstance(true_coeffs, torch.Tensor):
            assert true_coeffs.ndim == 1 and true_coeffs.numel() == len_vars
            self.true_coeffs = th.Variable(tensor=true_coeffs.unsqueeze(0))
        else:
            self.true_coeffs = true_coeffs

        batch_size = point.shape[0]
        if isinstance(point, torch.Tensor):
            if point.ndim == 1:
                point = point.unsqueeze(0)
            assert point.ndim == 2 and point.shape[1] == len_vars - 1
            self.point = th.Variable(
                tensor=torch.cat([point, torch.ones(batch_size, 1)], dim=1)
            )
        else:
            self.point = point
        for i, var in enumerate(optim_vars):
            attr_name = f"optim_var_{i}"
            setattr(self, attr_name, var)
            self.register_optim_var(attr_name)
        self.register_aux_var("true_coeffs")
        self.register_aux_var("point")

        if target is None:
            target_data = (self.point.tensor * self.true_coeffs.tensor).sum(
                1, keepdim=True
            ) ** 2
            if noise_mag:
                target_data += noise_mag * torch.randn(size=target_data.shape)
            self.target = th.Variable(tensor=target_data)
        else:
            self.target = target
        self.register_aux_var("target")

        self.noise_mag = noise_mag
        self.multivar = multivar

    def _eval_coeffs(self):
        if self.multivar:
            coeffs = torch.cat([v.tensor for v in self.optim_vars], axis=1)
        else:
            coeffs = self.optim_var_0.tensor
        return (self.point.tensor * coeffs).sum(1, keepdim=True)

    def error(self):
        # h(B * x) - h(Btrue * x)
        return self._eval_coeffs() ** 2 - self.target.tensor

    def jacobians(self):
        dhdz = 2 * self._eval_coeffs()
        grad = self.point.tensor * dhdz
        return [grad.unsqueeze(1)], self.error()

    def dim(self):
        return 1

    def _copy_impl(self, new_name=None):
        return ResidualCostFunction(
            [v.copy() for v in self._optim_vars],
            self.weight.copy(),
            name=new_name,
            true_coeffs=self.true_coeffs.copy(),
            point=self.point.copy(),
            multivar=self.multivar,
            noise_mag=self.noise_mag,
            target=self.target.copy(),
        )


def _check_info(info, batch_size, max_iterations, initial_error, objective):
    assert info.err_history.shape == (batch_size, max_iterations + 1)
    assert info.err_history[:, 0].allclose(initial_error)
    assert info.err_history.argmin(dim=1).allclose(info.best_iter + 1)
    last_error = objective.error_squared_norm() / 2
    last_convergence_idx = info.converged_iter.max().item()
    assert info.err_history[:, last_convergence_idx].allclose(last_error)


# This test uses least-squares regression to find the coefficients b_j of
#
#       y = f(x, B) = h(sum_i b_i * x_i + b_0)  ,  for i in [1, nvars-1]
#
# where h(z) = z^2.
#
# There is only one nvars-dimensional Vector for all coefficients, and
# one cost function per data-point (generated at random), which computes residual
#
#       (f(p, Btrue) - f(p, B))
#
# for some data point p specific to this cost function. There is no noise.
def _check_nonlinear_least_squares_fit(
    nonlinear_optim_cls,
    optimize_kwargs,
    points,
    nvars,
    npoints,
    batch_size,
):
    true_coeffs = torch.ones(nvars)
    variables = [th.Vector(nvars, name="coefficients")]
    cost_weight = th.ScaleCostWeight(1.0)
    objective = th.Objective()
    for i in range(npoints):
        objective.add(
            ResidualCostFunction(
                variables,
                cost_weight,
                true_coeffs=true_coeffs,
                point=points[i],
                name=f"residual_point_{i}",
            )
        )

    # Initial value is B = [0, 1, ..., nvars - 1]
    values = {"coefficients": torch.arange(nvars).repeat(batch_size, 1).float()}
    objective.update(values)
    initial_error = objective.error_squared_norm() / 2
    max_iterations = 20
    optimizer = nonlinear_optim_cls(objective)
    assert isinstance(optimizer.linear_solver, th.CholeskyDenseSolver)
    optimizer.set_params(max_iterations=max_iterations)
    info = optimizer.optimize(
        track_best_solution=True, track_err_history=True, **optimize_kwargs
    )
    # Solution must now match the true coefficients
    assert variables[0].tensor.allclose(true_coeffs.repeat(batch_size, 1), atol=1e-6)
    _check_info(info, batch_size, max_iterations, initial_error, objective)


# Same test as above but now each cost function has m 1-d variables instead
# of a single m-dim variable
def _check_nonlinear_least_squares_fit_multivar(
    nonlinear_optim_cls,
    optimize_kwargs,
    points,
    nvars,
    npoints,
    batch_size,
):
    true_coeffs = torch.ones(nvars)
    variables = [th.Vector(1, name=f"coeff{i}") for i in range(nvars)]
    cost_weight = th.ScaleCostWeight(1.0)
    objective = th.Objective()
    for i in range(npoints):
        objective.add(
            ResidualCostFunction(
                variables,
                cost_weight,
                true_coeffs=true_coeffs,
                point=points[i],
                name=f"residual_point_{i}",
                multivar=True,
            )
        )
    # Initial value is B = [0, 1, ..., nvars - 1]
    values = dict((f"coeff{i}", i * torch.ones(batch_size, 1)) for i in range(nvars))
    objective.update(values)
    initial_error = objective.error_squared_norm() / 2

    max_iterations = 20
    optimizer = nonlinear_optim_cls(objective)
    assert isinstance(optimizer.linear_solver, th.CholeskyDenseSolver)
    optimizer.set_params(max_iterations=max_iterations)
    info = optimizer.optimize(
        track_best_solution=True, track_err_history=True, **optimize_kwargs
    )

    # Solution must now match the true coefficients
    for i in range(nvars):
        assert variables[i].tensor.allclose(true_coeffs[i].repeat(batch_size, 1))

    _check_info(info, batch_size, max_iterations, initial_error, objective)


def _check_optimizer_returns_fail_status_on_singular(
    nonlinear_optim_cls, optimize_kwargs
):
    class BadLinearization(th.optimizer.Linearization):
        def __init__(self, objective):
            super().__init__(objective)

        def _linearize_jacobian_impl(self):
            pass

        def _linearize_hessian_impl(self):
            self.AtA = torch.zeros(
                self.objective.batch_size, self.num_rows, self.num_cols
            )
            self.Atb = torch.ones(self.objective.batch_size, self.num_cols)

    class MockCostFunction(th.CostFunction):
        def __init__(self, optim_vars, cost_weight):
            super().__init__(cost_weight)
            for var in optim_vars:
                setattr(self, var.name, var)
                self.register_optim_var(var.name)

        def error(self):
            return torch.ones(1)

        def jacobians(self):
            return torch.ones(1), self.error()

        def dim(self):
            return 1

        def _copy_impl(self):
            raise NotImplementedError

    objective = th.Objective()
    variables = [th.Vector(1, name="dummy")]
    objective.add(MockCostFunction(variables, th.ScaleCostWeight(1.0)))
    values = {"dummy": torch.zeros(1, 1)}
    objective.update(values)

    optimizer = nonlinear_optim_cls(objective, vectorize=False)
    assert isinstance(optimizer.linear_solver, th.CholeskyDenseSolver)
    optimizer.set_params(max_iterations=30)
    optimizer.linear_solver.linearization = BadLinearization(objective)
    with pytest.raises(RuntimeError):
        optimizer.optimize(track_best_solution=True, verbose=True, **optimize_kwargs)

    with pytest.warns(RuntimeWarning):
        with torch.no_grad():
            info = optimizer.optimize(
                track_best_solution=True, track_err_history=True, **optimize_kwargs
            )
        assert (info.status == th.NonlinearOptimizerStatus.FAIL).all()


def run_nonlinear_least_squares_check(
    nonlinear_optim_cls, optimize_kwargs, singular_check=True
):
    rng = torch.Generator()
    rng.manual_seed(0)
    nvars = 5
    npoints = 50
    batch_size = 32
    points = [torch.randn(batch_size, nvars - 1, generator=rng) for _ in range(npoints)]

    print("----- Check single variable formulation -----")
    _check_nonlinear_least_squares_fit(
        nonlinear_optim_cls,
        optimize_kwargs,
        points,
        nvars,
        npoints,
        batch_size,
    )
    print("----- Check multiple variable formulation -----")
    _check_nonlinear_least_squares_fit_multivar(
        nonlinear_optim_cls,
        optimize_kwargs,
        points,
        nvars,
        npoints,
        batch_size,
    )
    if singular_check:
        print("----- Check failure status on singular system formulation -----")
        _check_optimizer_returns_fail_status_on_singular(
            nonlinear_optim_cls, optimize_kwargs
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from itertools import count
from typing import Generator, Iterable, List, Optional, Sequence

from theseus.geometry import Manifold

from .variable import Variable


# Base class that provides objective-tracking functionality for functions of tensors
# such as cost fuctions and cost weights. Variables are classified as either
# optimization variables (of the objective), or auxiliary variables.
#
# Subclasses must register their optimization variables and auxiliary variables
# with methods `register_optim_vars` and `register_aux_vars`, respectively.
class TheseusFunction(abc.ABC):
    _ids = count(0)

    def __init__(self, name: Optional[str] = None):
        self._id = next(self._ids)
        if name:
            self.name = name
        else:
            self.name = self.get_default_name()
        self._optim_vars_attr_names: List[str] = []
        self._aux_vars_attr_names: List[str] = []

    def get_default_name(self) -> str:
        return f"{self.__class__.__name__}__{self._id}"

    def register_optim_var(self, attr_name: str):
        if attr_name in self._optim_vars_attr_names:
            raise ValueError(
                "Tried to register an optimization variable that was previously "
                "registered."
            )
        if attr_name in self._aux_vars_attr_names:
            raise ValueError(
                "An optimization variable cannot also be an auxiliary variable."
            )
        self._optim_vars_attr_names.append(attr_name)

    def register_aux_var(self, attr_name: str):
        if attr_name in self._aux_vars_attr_names:
            raise ValueError(
                "Tried to register an auxiliary variable that was previously registered."
            )
        if attr_name in self._optim_vars_attr_names:
            raise ValueError(
                "An auxiliary variable cannot also be an optimization variable."
            )
        self._aux_vars_attr_names.append(attr_name)

    def register_optim_vars(self, variable_names: Sequence[str]):
        for name in variable_names:
            self.register_optim_var(name)

    def register_aux_vars(self, aux_var_names: Sequence[str]):
        for name in aux_var_names:
            self.register_aux_var(name)

    def register_vars(self, vars: Iterable[Variable], is_optim_vars: bool = False):
        for var in vars:
            if hasattr(self, var.name):
                raise RuntimeError(f"Variable name {var.name} is not allowed.")
            setattr(self, var.name, var)
            if is_optim_vars:
                self.register_optim_var(var.name)
            else:
                self.register_aux_var(var.name)

    # Must copy everything
    @abc.abstractmethod
    def _copy_impl(self, new_name: Optional[str] = None) -> "TheseusFunction":
        pass

    def _has_duplicate_vars(self, another: "TheseusFunction") -> bool:
        def _check(base: Iterable[Variable], vectorized: Iterable[Variable]) -> bool:
            return len(set(base) & set(vectorized)) > 0

        return _check(self.optim_vars, another.optim_vars) | _check(
            self.aux_vars, another.aux_vars
        )

    def copy(
        self, new_name: Optional[str] = None, keep_variable_names: bool = False
    ) -> "TheseusFunction":
        if not new_name:
            new_name = f"{self.name}_copy"
        new_fn = self._copy_impl(new_name=new_name)
        if keep_variable_names:
            for old_var, new_var in zip(self.optim_vars, new_fn.optim_vars):
                new_var.name = old_var.name
            for old_aux, new_aux in zip(self.aux_vars, new_fn.aux_vars):
                new_aux.name = old_aux.name

        if self._has_duplicate_vars(new_fn):
            raise RuntimeError(
                f"{self.__class__.__name__}.copy() resulted in one of the original "
                "variables being reused. copy() requires all variables to be copied "
                "to new variables, so please re-implement _copy_impl() to satisfy this "
                "property."
            )

        return new_fn

    def __deepcopy__(self, memo):
        if id(self) in memo:
            return memo[id(self)]
        the_copy = self.copy()
        memo[id(self)] = the_copy
        return the_copy

    @property
    def optim_vars(self) -> Generator[Manifold, None, None]:
        return (getattr(self, attr) for attr in self._optim_vars_attr_names)

    @property
    def aux_vars(self) -> Generator[Variable, None, None]:
        return (getattr(self, attr) for attr in self._aux_vars_attr_names)

    def optim_var_at(self, index: int) -> Manifold:
        return getattr(self, self._optim_vars_attr_names[index])

    def aux_var_at(self, index: int) -> Variable:
        return getattr(self, self._aux_vars_attr_names[index])

    # This method is only used when copying the full objective.
    # It replaces the variable at the given index in the interal list of
    # variables with the variable passed as argument.
    # This is used by methods `set_optim_var_at()` and `set_aux_var_at()`.
    def _set_variable_at(
        self,
        index: int,
        variable: Variable,
        variable_attrs_array: List,
        variable_type: str,
    ):
        if index >= len(variable_attrs_array):
            raise ValueError(f"{variable_type} index out of range.")
        setattr(self, variable_attrs_array[index], variable)

    # see `_set_variable_at()`
    def set_optim_var_at(self, index: int, variable: Manifold):
        self._set_variable_at(
            index, variable, self._optim_vars_attr_names, "Optimization variable"
        )

    # see `_set_variable_at()`
    def set_aux_var_at(self, index: int, variable: Variable):
        self._set_variable_at(
            index, variable, self._aux_vars_attr_names, "Auxiliary variable"
        )

    def num_optim_vars(self):
        return len(self._optim_vars_attr_names)

    def num_aux_vars(self):
        return len(self._aux_vars_attr_names)

    # calls to() on the cost weight, optimization variables and any internal tensors
    def to(self, *args, **kwargs):
        for var in self.optim_vars:
            var.to(*args, **kwargs)
        for aux in self.aux_vars:
            aux.to(*args, **kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Optional, Type

import torch

from theseus.core import Objective
from theseus.optimizer import Linearization
from theseus.optimizer.linear import LinearSolver

from .nonlinear_least_squares import NonlinearLeastSquares


class GaussNewton(NonlinearLeastSquares):
    def __init__(
        self,
        objective: Objective,
        linear_solver_cls: Optional[Type[LinearSolver]] = None,
        vectorize: bool = True,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        linear_solver_kwargs: Optional[Dict[str, Any]] = None,
        abs_err_tolerance: float = 1e-10,
        rel_err_tolerance: float = 1e-8,
        max_iterations: int = 20,
        step_size: float = 1.0,
        **kwargs
    ):
        super().__init__(
            objective,
            linear_solver_cls=linear_solver_cls,
            vectorize=vectorize,
            linearization_cls=linearization_cls,
            linearization_kwargs=linearization_kwargs,
            linear_solver_kwargs=linear_solver_kwargs,
            abs_err_tolerance=abs_err_tolerance,
            rel_err_tolerance=rel_err_tolerance,
            max_iterations=max_iterations,
            step_size=step_size,
            **kwargs
        )

    def compute_delta(self, **kwargs) -> torch.Tensor:
        return self.linear_solver.solve()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
from scipy.sparse import csr_matrix, lil_matrix


def random_sparse_binary_matrix(rows, cols, fill, min_entries_per_col) -> csr_matrix:
    retv = lil_matrix((rows, cols))

    if min_entries_per_col > 0:
        min_entries_per_col = min(rows, min_entries_per_col)
        rows_array = np.arange(rows)
        for c in range(cols):
            for r in np.random.choice(rows_array, min_entries_per_col):
                retv[r, c] = 1.0

    num_entries = int(fill * rows * cols)
    while retv.getnnz() < num_entries:
        col = np.random.randint(cols)
        row = np.random.randint(rows)
        retv[row, col] = 1.0

    return retv.tocsr()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

from theseus.core import CostFunction, CostWeight, Variable
from theseus.embodied.kinematics import IdentityModel, KinematicsModel
from theseus.geometry import Point2

from .signed_distance_field import SignedDistanceField2D


class Collision2D(CostFunction):
    def __init__(
        self,
        pose: Point2,
        sdf_origin: Variable,
        sdf_data: Variable,
        sdf_cell_size: Variable,
        cost_eps: Union[float, Variable, torch.Tensor],
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        if not isinstance(pose, Point2):
            raise ValueError("Collision2D only accepts 2D poses as inputs.")
        super().__init__(cost_weight, name=name)
        self.pose = pose
        self.sdf_origin = sdf_origin
        self.sdf_data = sdf_data
        self.sdf_cell_size = sdf_cell_size
        if not isinstance(cost_eps, Variable):
            if not isinstance(cost_eps, torch.Tensor):
                cost_eps = torch.tensor(cost_eps)
            self.cost_eps = Variable(cost_eps)
        else:
            self.cost_eps = cost_eps
        self.cost_eps.tensor = self.cost_eps.tensor.view(-1, 1)
        self.register_optim_vars(["pose"])
        self.register_aux_vars(["sdf_origin", "sdf_data", "sdf_cell_size", "cost_eps"])
        self.robot: KinematicsModel = IdentityModel()
        self.sdf = SignedDistanceField2D(sdf_origin, sdf_cell_size, sdf_data)

    def _compute_distances_and_jacobians(
        self,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        robot_state = cast(Point2, self.robot.forward_kinematics(self.pose)["state"])
        return self.sdf.signed_distance(robot_state.tensor.view(-1, 2, 1))

    def _error_from_distances(self, distances: torch.Tensor):
        return (self.cost_eps.tensor - distances).clamp(min=0)

    def error(self) -> torch.Tensor:
        distances, _ = self._compute_distances_and_jacobians()
        return self._error_from_distances(distances)

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        distances, jacobian = self._compute_distances_and_jacobians()
        error = self._error_from_distances(distances)
        faraway_idx = distances > self.cost_eps.tensor
        jacobian[faraway_idx] = 0.0
        return [-jacobian], error

    def _copy_impl(self, new_name: Optional[str] = None) -> "Collision2D":
        return Collision2D(
            self.pose.copy(),
            self.sdf_origin.copy(),
            self.sdf_data.copy(),
            self.sdf_cell_size.copy(),
            self.cost_eps.copy(),
            self.weight.copy(),
            name=new_name,
        )

    def dim(self) -> int:
        return 1

    # This is needed so that the SDF container also updates with the new aux var
    def set_aux_var_at(self, index: int, variable: Variable):
        super().set_aux_var_at(index, variable)
        self.sdf.update_data(self.sdf_origin, self.sdf_data, self.sdf_cell_size)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, cast

import numpy as np
import torch

import theseus as th

from .util import random_small_quaternion


class Camera:
    def __init__(
        self,
        pose: th.SE3,
        focal_length: th.Vector,
        calib_k1: th.Vector,
        calib_k2: th.Vector,
    ):
        self.pose = pose
        self.focal_length = focal_length
        self.calib_k1 = calib_k1
        self.calib_k2 = calib_k2

    def to_params(self) -> List[float]:
        batch_size = self.pose.tensor.shape[0]
        assert batch_size == 1
        R = th.SO3(tensor=self.pose.tensor[:, :, :3]).log_map().squeeze(0)
        t = self.pose.tensor[:, :, 3].squeeze(0)
        return [
            *R.numpy(),
            *t.numpy(),
            float(self.focal_length[0, 0]),
            float(self.calib_k1[0, 0]),
            float(self.calib_k2[0, 0]),
        ]

    @staticmethod
    def from_params(params: List[float], name: str = "Cam") -> "Camera":
        r = th.SO3.exp_map(torch.tensor(params[:3], dtype=torch.float64).unsqueeze(0))
        t = torch.tensor([params[3:6]], dtype=torch.float64).unsqueeze(2)
        pose = th.SE3(tensor=torch.cat([r.tensor, t], dim=2), name=name + "_pose")
        focal_length = th.Vector(
            tensor=torch.tensor([params[6:7]], dtype=torch.float64),
            name=name + "_focal_length",
        )
        calib_k1 = th.Vector(
            tensor=torch.tensor([params[7:8]], dtype=torch.float64),
            name=name + "_calib_k1",
        )
        calib_k2 = th.Vector(
            tensor=torch.tensor([params[8:9]], dtype=torch.float64),
            name=name + "_calib_k2",
        )
        return Camera(pose, focal_length, calib_k1, calib_k2)

    def project_point(self, point: th.Point3) -> torch.Tensor:
        point_cam = self.pose.transform_from(point)
        proj = -point_cam.tensor[:, :2] / point_cam.tensor[:, 2:3]
        proj_sqn = (proj * proj).sum(dim=1).unsqueeze(1)
        proj_factor = self.focal_length.tensor * (
            1.0 + proj_sqn * (self.calib_k1.tensor + proj_sqn * self.calib_k2.tensor)
        )
        return proj * proj_factor

    def perturbed(self, rot_random: float = 0.1, pos_random: float = 0.2):
        batch_size = self.pose.shape[0]
        pert_rot = torch.cat(
            [
                random_small_quaternion(max_degrees=rot_random).unsqueeze(0)
                for _ in range(batch_size)
            ]
        )
        pert_tr = (
            torch.rand((batch_size, 3), dtype=torch.float64) * 2
            + torch.tensor([-1, -1, -1], dtype=torch.float64)
        ) * pos_random
        pert_data = torch.cat([pert_tr, pert_rot], dim=1)
        pert = th.SE3(pert_data)
        pert_pose = cast(th.SE3, pert.compose(self.pose))
        return Camera(
            pert_pose.copy(new_name=self.pose.name + "_copy"),
            self.focal_length.copy(new_name=self.focal_length.name + "_copy"),
            self.calib_k1.copy(new_name=self.calib_k1.name + "_copy"),
            self.calib_k2.copy(new_name=self.calib_k2.name + "_copy"),
        )

    @staticmethod
    def generate_synthetic(
        batch_size: int = 1,
        rot_random: float = 20.0,
        pos_random: float = 1.0,
        pos_base: Optional[torch.Tensor] = None,
        fl_random: float = 100.0,
        fl_base: float = 1000.0,
        k1_random: float = 0.1,
        k1_base: float = 0.0,
        k2_random: float = 0.05,
        k2_base: float = 0.0,
        name: str = "Cam",
    ):
        if pos_base is None:
            pos_base = torch.zeros(3, dtype=torch.float64)
        cam_rot = torch.cat(
            [
                random_small_quaternion(max_degrees=rot_random).unsqueeze(0)
                for _ in range(batch_size)
            ]
        )
        cam_tr = (
            torch.rand((batch_size, 3), dtype=torch.float64) * 2
            + torch.tensor([-1, -1, -1], dtype=torch.float64)
        ) * pos_random + pos_base
        cam_pose_data = torch.cat([cam_tr, cam_rot], dim=1)
        cam_pose = th.SE3(cam_pose_data, name=name + "_pose")

        focal_length = th.Vector(
            tensor=(torch.rand((batch_size, 1), dtype=torch.float64) * 2 - 1.0)
            * fl_random
            + fl_base,
            name=name + "_focal_length",
        )
        calib_k1 = th.Vector(
            tensor=(torch.rand((batch_size, 1), dtype=torch.float64) * 2 - 1.0)
            * k1_random
            + k1_base,
            name=name + "_calib_k1",
        )
        calib_k2 = th.Vector(
            tensor=(torch.rand((batch_size, 1), dtype=torch.float64) * 2 - 1.0)
            * k2_random
            + k2_base,
            name=name + "_calib_k2",
        )
        return Camera(cam_pose, focal_length, calib_k1, calib_k2)


class Observation:
    def __init__(
        self, camera_index: int, point_index: int, image_feature_point: th.Point2
    ):
        self.camera_index = camera_index
        self.point_index = point_index
        self.image_feature_point = image_feature_point


class BundleAdjustmentDataset:
    def __init__(
        self,
        cameras: List[Camera],
        points: List[th.Point3],
        observations: List[Observation],
        gt_cameras: Optional[List[Camera]] = None,
        gt_points: Optional[List[th.Point3]] = None,
    ):
        self.cameras = cameras
        self.points = points
        self.observations = observations
        self.gt_cameras = gt_cameras
        self.gt_points = gt_points

    @staticmethod
    def load_bal_dataset(path: str):
        observations = []
        cameras = []
        points = []
        with open(path, "rt") as out:
            num_cameras, num_points, num_observations = [
                int(x) for x in out.readline().rstrip().split()
            ]
            for i in range(num_observations):
                fields = out.readline().rstrip().split()
                feat = th.Point2(
                    tensor=torch.tensor(
                        [float(fields[2]), float(fields[3])], dtype=torch.float64
                    ).unsqueeze(0),
                    name=f"Feat{i}",
                )
                observations.append(
                    Observation(
                        camera_index=int(fields[0]),
                        point_index=int(fields[1]),
                        image_feature_point=feat,
                    )
                )

            for i in range(num_cameras):
                params = []
                for _ in range(9):
                    params.append(float(out.readline().rstrip()))
                cameras.append(Camera.from_params(params, name=f"Cam{i}"))

            for i in range(num_points):
                params = []
                for _ in range(3):
                    params.append(float(out.readline().rstrip()))
                points.append(
                    th.Point3(
                        tensor=torch.tensor(params, dtype=torch.float64).unsqueeze(0),
                        name=f"Pt{i}",
                    )
                )
        return cameras, points, observations

    @staticmethod
    def save_bal_dataset(
        path: str,
        cameras: List[Camera],
        points: List[th.Point3],
        observations: List[Observation],
    ):
        with open(path, "wt") as out:
            print(f"{len(cameras)} {len(points)} {len(observations)}", file=out)
            for obs in observations:
                f = obs.image_feature_point.tensor.squeeze(0).numpy()
                print(f"{obs.camera_index} {obs.point_index} {f[0]} {f[1]}", file=out)
            for cam in cameras:
                params = cam.to_params()
                for p in params:
                    print(f"{p}", file=out)
            for pt in points:
                params = pt.tensor.squeeze(0).numpy()
                for p in params:
                    print(f"{p}", file=out)

    @staticmethod
    def load_from_file(path, gt_path: Optional[str] = None):
        cameras, points, observations = BundleAdjustmentDataset.load_bal_dataset(path)
        if gt_path is not None:
            gt_points, gt_cameras, _ = BundleAdjustmentDataset.load_bal_dataset(gt_path)
        else:
            gt_points, gt_cameras = None, None
        return BundleAdjustmentDataset(
            cameras=cameras,
            points=points,
            observations=observations,
            gt_cameras=gt_cameras,
            gt_points=gt_points,
        )

    def save_to_file(self, path, gt_path: Optional[str] = None):
        BundleAdjustmentDataset.save_bal_dataset(
            path, self.cameras, self.points, self.observations
        )
        if gt_path is not None:
            BundleAdjustmentDataset.save_bal_dataset(
                gt_path, self.gt_cameras, self.gt_points, self.observations
            )

    def histogram(self) -> str:
        buckets = np.zeros(11)
        for obs in self.observations:
            proj_pt = self.cameras[obs.camera_index].project_point(
                self.points[obs.point_index]
            )
            error = float((obs.image_feature_point.tensor - proj_pt).norm())
            idx = min(int(error), len(buckets) - 1)
            buckets[idx] += 1
        max_buckets = max(buckets)
        hist_str = ""
        for i in range(len(buckets)):
            bi = buckets[i]
            label = f"{i}-{i+1}" if i + 1 < len(buckets) else f"{i}+"
            barlen = round(bi * 80 / max_buckets)
            hist_str += f"{label}: {'#' * barlen} {bi}\n"
        return hist_str

    @staticmethod
    def generate_synthetic(
        num_cameras: int,
        num_points: int,
        average_track_length: int = 7,
        track_locality: float = 0.1,
        feat_random: float = 1.5,
        prob_feat_is_outlier: float = 0.02,
        outlier_feat_random: float = 70,
    ):

        # add cameras
        gt_cameras = [
            Camera.generate_synthetic(
                pos_base=torch.tensor(
                    [-i * 100.0 / (num_cameras - 1), 0, -100], dtype=torch.float64
                ),
                name=f"Cam{i}",
            )
            for i in range(num_cameras)
        ]
        cameras = [cam.perturbed() for cam in gt_cameras]

        # add points
        gt_points = [
            th.Point3(
                tensor=(torch.rand((1, 3), dtype=torch.float64) * 2 - 1) * 20
                + torch.tensor([i * 100.0 / (num_points), 0, 0], dtype=torch.float64),
                name=f"Pt{i}",
            )
            for i in range(num_points)
        ]
        points = [
            th.Point3(
                tensor=gt_points[i].tensor + (torch.rand((1, 3)) * 2 - 1) * 0.2,
                name=gt_points[i].name + "_copy",
            )
            for i in range(num_points)
        ]

        # add observations
        pts_per_cam = average_track_length * num_points // num_cameras
        observations = []
        obs_num = 0
        for i in range(num_cameras):
            span_size = min(pts_per_cam + int(track_locality * num_points), num_points)
            span_start = (num_points - span_size) * i // num_cameras
            obs_pts = torch.randperm(span_size)[:pts_per_cam] + span_start
            for j in obs_pts:
                feat = (
                    gt_cameras[i].project_point(gt_points[j])
                    + (torch.rand(1, 2) * 2 - 1) * feat_random
                )
                if torch.rand(()) < prob_feat_is_outlier:
                    feat = feat + (torch.rand(1, 2) * 2 - 1) * outlier_feat_random
                observations.append(
                    Observation(
                        camera_index=i,
                        point_index=j,
                        image_feature_point=th.Point2(
                            tensor=feat, name=f"Feat{obs_num}"
                        ),
                    )
                )
                obs_num += 1

        return BundleAdjustmentDataset(
            cameras, points, observations, gt_cameras, gt_points
        )


def ba_histogram(
    cameras: List[Camera],
    points: List[th.Point3],
    observations: List[Observation],
) -> str:
    ba = BundleAdjustmentDataset(cameras, points, observations)
    return ba.histogram()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.constants import EPS
from theseus.core.tests.common import check_copy_var

from .common import (
    check_jacobian_for_local,
    check_projection_for_exp_map,
    check_projection_for_log_map,
)

torch.manual_seed(0)


def test_item():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            assert torch.allclose(v1.tensor, t1)
            v1[0, 0] = 11.1
            assert not torch.allclose(v1.tensor, t1)
            t1[0, 0] = 11.1
            assert torch.allclose(v1.tensor, t1)


def test_add():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            t2 = torch.rand(i, j)
            v2 = th.Vector(tensor=t2.clone(), name="v2")
            vsum = th.Vector(tensor=t1 + t2)
            assert (v1 + v2).allclose(vsum)
            assert v1.compose(v2).allclose(vsum)


def test_sub():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            t2 = torch.rand(i, j)
            v2 = th.Vector(tensor=t2.clone(), name="v2")
            assert (v1 - v2).allclose(th.Vector(tensor=t1 - t2))
            v2 = -v2
            assert (v1 + v2).allclose(th.Vector(tensor=t1 - t2))


def test_mul():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            assert (v1 * torch.tensor(2.1)).allclose(th.Vector(tensor=t1 * 2.1))
            assert (torch.tensor(1.1) * v1).allclose(th.Vector(tensor=t1 * 1.1))


def test_div():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            assert (v1 / torch.tensor(2.1)).allclose(th.Vector(tensor=t1 / 2.1))


def test_matmul():
    for i in range(1, 4):
        for j in range(1, 4):
            for k in range(1, 4):
                t = torch.rand(i, j, k)
                t1 = torch.rand(i, j)
                v1 = th.Vector(tensor=t1.clone(), name="v1")
                v1t = v1 @ t
                assert v1t.allclose((t1.unsqueeze(1) @ t).squeeze(1))
                assert v1t.shape == (i, k)
                t2 = torch.rand(i, k)
                v2 = th.Vector(tensor=t2.clone(), name="v2")
                tv2 = t @ v2
                assert tv2.allclose((t @ t2.unsqueeze(2)).squeeze(2))
                assert tv2.shape == (i, j)


def test_dot():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            t2 = torch.rand(i, j)
            v2 = th.Vector(tensor=t2.clone(), name="v2")
            assert torch.allclose(v1.dot(v2), torch.mul(t1, t2).sum(-1))
            assert torch.allclose(v1.inner(v2), torch.mul(t1, t2).sum(-1))


def test_outer():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            t2 = torch.rand(i, j)
            v2 = th.Vector(tensor=t2.clone(), name="v2")
            assert torch.allclose(
                v1.outer(v2), torch.matmul(t1.unsqueeze(2), t2.unsqueeze(1))
            )


def test_abs():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            assert v1.abs().allclose(th.Vector(tensor=t1.abs()))


def test_norm():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            assert v1.norm() == t1.norm()
            assert v1.norm(p="fro") == torch.norm(t1, p="fro")


def test_cat():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            t2 = torch.rand(i, j)
            v2 = th.Vector(tensor=t2.clone(), name="v2")
            t3 = torch.rand(i, j)
            v3 = th.Vector(tensor=t3.clone(), name="v3")
            assert v1.cat(v2).allclose(th.Vector(tensor=torch.cat((t1, t2), 1)))
            assert v1.cat((v2, v3)).allclose(
                th.Vector(tensor=torch.cat((t1, t2, t3), 1))
            )


def test_local():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            t2 = torch.rand(i, j)
            v2 = th.Vector(tensor=t2.clone(), name="v2")
            assert torch.allclose(v1._local_impl(v2), t2 - t1)
            assert torch.allclose(v1.local(v2), t2 - t1)
            assert torch.allclose(th.local(v1, v2), t2 - t1)


def test_retract():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v1 = th.Vector(tensor=t1.clone(), name="v1")
            d = torch.rand(i, j)
            assert v1._retract_impl(d).allclose(th.Vector(tensor=t1 + d))
            assert v1.retract(d).allclose(th.Vector(tensor=t1 + d))
            assert th.retract(v1, d).allclose(th.Vector(tensor=t1 + d))


def test_update():
    for dof in range(1, 21):
        v = th.Vector(dof)
        for _ in range(100):
            rng = np.random.default_rng()
            batch_size = rng.integers(low=1, high=100)
            data = torch.rand(batch_size, dof)
            v.update(data)
            assert torch.allclose(data, v.tensor)


def test_copy():
    for i in range(1, 4):
        for j in range(1, 5):
            t1 = torch.rand(i, j)
            v = th.Vector(tensor=t1.clone(), name="v")
            check_copy_var(v)


def test_exp_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        dim = torch.randint(1, 10, size=[1], generator=rng)[0]
        tangent_vector = torch.rand(batch_size, dim, generator=rng).double() - 0.5
        ret = th.Vector.exp_map(tangent_vector)

        assert torch.allclose(ret.tensor, tangent_vector, atol=EPS)
        check_projection_for_exp_map(
            tangent_vector, Group=th.Vector, is_projected=False
        )


def test_log_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        dim = torch.randint(1, 10, size=[1], generator=rng)[0]
        group = th.Vector.rand(batch_size, dim, generator=rng)
        ret = group.log_map()

        assert torch.allclose(ret, group.tensor, atol=EPS)
        check_projection_for_log_map(
            tangent_vector=ret, Group=th.Vector, is_projected=False
        )


def test_local_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        dim = torch.randint(1, 10, size=[1], generator=rng)[0]
        group0 = th.Vector.rand(batch_size, dim, generator=rng)
        group1 = th.Vector.rand(batch_size, dim, generator=rng)

        check_jacobian_for_local(group0, group1, Group=th.Vector, is_projected=False)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .lie_group import LieGroup, adjoint, between, compose, exp_map, inverse, log_map
from .manifold import Manifold, OptionalJacobians, local, retract
from .point_types import (
    Point2,
    Point3,
    rand_point2,
    rand_point3,
    randn_point2,
    randn_point3,
)
from .se2 import SE2, rand_se2, randn_se2
from .se3 import SE3, rand_se3, randn_se3
from .so2 import SO2, rand_so2, randn_so2
from .so3 import SO3, rand_so3, randn_so3
from .utils import (
    LieGroupTensor,
    enable_lie_tangent,
    no_lie_tangent,
    set_lie_tangent_enabled,
)
from .vector import Vector, rand_vector, randn_vector



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import math
import warnings
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Type

import numpy as np
import torch

import theseus.constants
from theseus.core import Objective
from theseus.optimizer import Linearization, Optimizer, OptimizerInfo
from theseus.optimizer.linear import LinearSolver


@dataclass
class NonlinearOptimizerParams:
    abs_err_tolerance: float
    rel_err_tolerance: float
    max_iterations: int
    step_size: float

    def update(self, params_dict):
        for param, value in params_dict.items():
            if hasattr(self, param):
                setattr(self, param, value)
            else:
                raise ValueError(f"Invalid nonlinear optimizer parameter {param}.")


class NonlinearOptimizerStatus(Enum):
    START = 0
    CONVERGED = 1
    MAX_ITERATIONS = 2
    FAIL = -1


# All info information is batched
@dataclass
class NonlinearOptimizerInfo(OptimizerInfo):
    converged_iter: torch.Tensor
    best_iter: torch.Tensor
    err_history: Optional[torch.Tensor]
    last_err: torch.Tensor
    best_err: torch.Tensor
    state_history: Optional[Dict[str, torch.Tensor]]  # variable name to state history


class BackwardMode(Enum):
    FULL = 0
    IMPLICIT = 1
    TRUNCATED = 2
    DLM = 3


class NonlinearOptimizer(Optimizer, abc.ABC):
    def __init__(
        self,
        objective: Objective,
        linear_solver_cls: Type[LinearSolver],
        *args,
        vectorize: bool = True,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        linear_solver_kwargs: Optional[Dict[str, Any]] = None,
        abs_err_tolerance: float = 1e-10,
        rel_err_tolerance: float = 1e-8,
        max_iterations: int = 20,
        step_size: float = 1.0,
        **kwargs,
    ):
        super().__init__(objective, vectorize=vectorize, **kwargs)
        linear_solver_kwargs = linear_solver_kwargs or {}
        self.linear_solver = linear_solver_cls(
            objective,
            linearization_cls=linearization_cls,
            linearization_kwargs=linearization_kwargs,
            **linear_solver_kwargs,
        )
        self.params = NonlinearOptimizerParams(
            abs_err_tolerance, rel_err_tolerance, max_iterations, step_size
        )

    def set_params(self, **kwargs):
        self.params.update(kwargs)

    def _check_convergence(self, err: torch.Tensor, last_err: torch.Tensor):
        assert not torch.is_grad_enabled()
        if err.abs().mean() < theseus.constants.EPS:
            return torch.ones_like(err).bool()

        abs_error = (last_err - err).abs()
        rel_error = abs_error / last_err
        return (abs_error < self.params.abs_err_tolerance).logical_or(
            rel_error < self.params.rel_err_tolerance
        )

    def _maybe_init_best_solution(
        self, do_init: bool = False
    ) -> Optional[Dict[str, torch.Tensor]]:
        if not do_init:
            return None
        solution_dict = {}
        for var in self.linear_solver.linearization.ordering:
            solution_dict[var.name] = var.tensor.detach().clone().cpu()
        return solution_dict

    def _init_info(
        self,
        track_best_solution: bool,
        track_err_history: bool,
        track_state_history: bool,
    ) -> NonlinearOptimizerInfo:
        with torch.no_grad():
            last_err = self.objective.error_squared_norm() / 2
        best_err = last_err.clone() if track_best_solution else None
        if track_err_history:
            err_history = (
                torch.ones(self.objective.batch_size, self.params.max_iterations + 1)
                * math.inf
            )
            assert last_err.grad_fn is None
            err_history[:, 0] = last_err.clone().cpu()
        else:
            err_history = None

        if track_state_history:
            state_history = {}
            for var in self.objective.optim_vars.values():
                state_history[var.name] = (
                    torch.ones(
                        self.objective.batch_size,
                        *var.shape[1:],
                        self.params.max_iterations + 1,
                    )
                    * math.inf
                )
                state_history[var.name][..., 0] = var.tensor.detach().clone().cpu()
        else:
            state_history = None

        return NonlinearOptimizerInfo(
            best_solution=self._maybe_init_best_solution(do_init=track_best_solution),
            last_err=last_err,
            best_err=best_err,
            status=np.array(
                [NonlinearOptimizerStatus.START] * self.objective.batch_size
            ),
            converged_iter=torch.zeros_like(last_err, dtype=torch.long),
            best_iter=torch.zeros_like(last_err, dtype=torch.long),
            err_history=err_history,
            state_history=state_history,
        )

    def _update_state_history(self, iter_idx: int, info: NonlinearOptimizerInfo):
        for var in self.objective.optim_vars.values():
            info.state_history[var.name][..., iter_idx + 1] = (
                var.tensor.detach().clone().cpu()
            )

    def _update_info(
        self,
        info: NonlinearOptimizerInfo,
        current_iter: int,
        err: torch.Tensor,
        converged_indices: torch.Tensor,
    ):
        info.converged_iter += 1 - converged_indices.long()
        if info.err_history is not None:
            assert err.grad_fn is None
            info.err_history[:, current_iter + 1] = err.clone().cpu()
        if info.state_history is not None:
            self._update_state_history(current_iter, info)

        if info.best_solution is not None:
            # Only copy best solution if needed (None means track_best_solution=False)
            assert info.best_err is not None
            good_indices = err < info.best_err
            info.best_iter[good_indices] = current_iter
            for var in self.linear_solver.linearization.ordering:
                info.best_solution[var.name][good_indices] = (
                    var.tensor.detach().clone()[good_indices].cpu()
                )

            info.best_err = torch.minimum(info.best_err, err)

        converged_indices = self._check_convergence(err, info.last_err)
        info.status[
            np.array(converged_indices.detach().cpu())
        ] = NonlinearOptimizerStatus.CONVERGED

    # Modifies the (no grad) info in place to add data of grad loop info
    def _merge_infos(
        self,
        grad_loop_info: NonlinearOptimizerInfo,
        num_no_grad_iters: int,
        num_grad_iters: int,
        info: NonlinearOptimizerInfo,
    ):
        total_iters = num_no_grad_iters + num_grad_iters
        # we add + 1 to all indices to account for the initial values
        info_idx = slice(num_no_grad_iters + 1, total_iters + 1)
        grad_info_idx = slice(1, num_grad_iters + 1)
        # Concatenate error histories
        if info.err_history is not None:
            info.err_history[:, info_idx] = grad_loop_info.err_history[:, grad_info_idx]
        if info.state_history is not None:
            for var in self.objective.optim_vars.values():
                info.state_history[var.name][
                    ..., info_idx
                ] = grad_loop_info.state_history[var.name][..., grad_info_idx]

        # Merge best solution and best error
        if info.best_solution is not None:
            best_solution = {}
            best_err_no_grad = info.best_err
            best_err_grad = grad_loop_info.best_err
            idx_no_grad = (best_err_no_grad < best_err_grad).cpu().view(-1, 1)
            best_err = torch.minimum(best_err_no_grad, best_err_grad)
            for var_name in info.best_solution:
                sol_no_grad = info.best_solution[var_name]
                sol_grad = grad_loop_info.best_solution[var_name]
                best_solution[var_name] = torch.where(
                    idx_no_grad, sol_no_grad, sol_grad
                )
            info.best_solution = best_solution
            info.best_err = best_err

        # Merge the converged status into the info from the detached loop,
        M = info.status == NonlinearOptimizerStatus.MAX_ITERATIONS
        assert np.all(
            (grad_loop_info.status[M] == NonlinearOptimizerStatus.MAX_ITERATIONS)
            | (grad_loop_info.status[M] == NonlinearOptimizerStatus.CONVERGED)
        )
        info.status[M] = grad_loop_info.status[M]
        info.converged_iter[M] = (
            info.converged_iter[M] + grad_loop_info.converged_iter[M]
        )
        # If didn't coverge in either loop, remove misleading converged_iter value
        info.converged_iter[
            M & (grad_loop_info.status == NonlinearOptimizerStatus.MAX_ITERATIONS)
        ] = -1

    # loop for the iterative optimizer
    def _optimize_loop(
        self,
        num_iter: int,
        info: NonlinearOptimizerInfo,
        verbose: bool,
        truncated_grad_loop: bool,
        **kwargs,
    ) -> int:
        converged_indices = torch.zeros_like(info.last_err).bool()
        iters_done = 0
        for it_ in range(num_iter):
            iters_done += 1
            # do optimizer step
            self.linear_solver.linearization.linearize()
            try:
                if truncated_grad_loop:
                    # The derivation for implicit differentiation states that
                    # the autograd-enabled loop (which `truncated_grad_loop` signals)
                    # must be done using Gauss-Newton steps. Well, technically,
                    # full Newton, but it seems less stable numerically and
                    # GN is working well so far.
                    delta = self.linear_solver.solve()
                else:
                    delta = self.compute_delta(**kwargs)
            except RuntimeError as run_err:
                msg = (
                    f"There was an error while running the linear optimizer. "
                    f"Original error message: {run_err}."
                )
                if torch.is_grad_enabled():
                    raise RuntimeError(
                        msg + " Backward pass will not work. To obtain "
                        "the best solution seen before the error, run with "
                        "torch.no_grad()"
                    )
                else:
                    warnings.warn(msg, RuntimeWarning)
                    info.status[:] = NonlinearOptimizerStatus.FAIL
                    return iters_done

            if truncated_grad_loop:
                # This is a "secret" option that is currently being tested in the
                # context of implicit differentiation. Might be added as a supported
                # kwarg in the future with a different name, or removed altogether.
                if "__keep_final_step_size__" in kwargs:
                    step_size = self.params.step_size
                else:
                    step_size = 1.0
                force_update = True
            else:
                step_size = self.params.step_size
                force_update = False

            self.objective.retract_optim_vars(
                delta * step_size,
                self.linear_solver.linearization.ordering,
                ignore_mask=converged_indices,
                force_update=force_update,
            )

            # check for convergence
            with torch.no_grad():
                err = self.objective.error_squared_norm() / 2
                self._update_info(info, it_, err, converged_indices)
                if verbose:
                    print(
                        f"Nonlinear optimizer. Iteration: {it_+1}. "
                        f"Error: {err.mean().item()}"
                    )
                converged_indices = self._check_convergence(err, info.last_err)
                info.status[
                    converged_indices.cpu().numpy()
                ] = NonlinearOptimizerStatus.CONVERGED
                if converged_indices.all():
                    break  # nothing else will happen at this point
                info.last_err = err

        info.status[
            info.status == NonlinearOptimizerStatus.START
        ] = NonlinearOptimizerStatus.MAX_ITERATIONS
        return iters_done

    # `track_best_solution` keeps a **detached** copy (as in no gradient info)
    # of the best variables found, but it is optional to avoid unnecessary copying
    # if this is not needed
    def _optimize_impl(
        self,
        track_best_solution: bool = False,
        track_err_history: bool = False,
        track_state_history: bool = False,
        verbose: bool = False,
        backward_mode: BackwardMode = BackwardMode.FULL,
        **kwargs,
    ) -> OptimizerInfo:
        with torch.no_grad():
            info = self._init_info(
                track_best_solution, track_err_history, track_state_history
            )

        if verbose:
            print(
                f"Nonlinear optimizer. Iteration: 0. "
                f"Error: {info.last_err.mean().item()}"
            )

        if backward_mode in [BackwardMode.FULL, BackwardMode.DLM]:
            self._optimize_loop(
                start_iter=0,
                num_iter=self.params.max_iterations,
                info=info,
                verbose=verbose,
                truncated_grad_loop=False,
                **kwargs,
            )
            # If didn't coverge, remove misleading converged_iter value
            info.converged_iter[
                info.status == NonlinearOptimizerStatus.MAX_ITERATIONS
            ] = -1
            return info
        elif backward_mode in [BackwardMode.IMPLICIT, BackwardMode.TRUNCATED]:
            if backward_mode == BackwardMode.IMPLICIT:
                backward_num_iterations = 1
            else:
                if "backward_num_iterations" not in kwargs:
                    raise ValueError(
                        "backward_num_iterations expected but not received"
                    )
                if kwargs["backward_num_iterations"] > self.params.max_iterations:
                    warnings.warn(
                        f"Input backward_num_iterations "
                        f"(={kwargs['backward_num_iterations']}) > "
                        f"max_iterations (={self.params.max_iterations}). "
                        f"Using backward_num_iterations=max_iterations."
                    )
                backward_num_iterations = min(
                    kwargs["backward_num_iterations"], self.params.max_iterations
                )

            num_no_grad_iter = self.params.max_iterations - backward_num_iterations
            with torch.no_grad():
                # actual_num_iters could be < num_iter due to early convergence
                no_grad_iters_done = self._optimize_loop(
                    num_iter=num_no_grad_iter,
                    info=info,
                    verbose=verbose,
                    truncated_grad_loop=False,
                    **kwargs,
                )

            grad_loop_info = self._init_info(
                track_best_solution, track_err_history, track_state_history
            )
            grad_iters_done = self._optimize_loop(
                num_iter=backward_num_iterations,
                info=grad_loop_info,
                verbose=verbose,
                truncated_grad_loop=True,
                **kwargs,
            )

            # Adds grad_loop_info results to original info
            self._merge_infos(grad_loop_info, no_grad_iters_done, grad_iters_done, info)

            return info
        else:
            raise ValueError("Unrecognized backward mode")

    @abc.abstractmethod
    def compute_delta(self, **kwargs) -> torch.Tensor:
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import pathlib
import random
from typing import Dict, List, Optional, Tuple

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.utils.data
import yaml

import theseus as th

# ----------------------------------------------------------------------------------- #
# ------------------------------------ Data loading --------------------------------- #
# ----------------------------------------------------------------------------------- #
FileInfo = Tuple[pathlib.Path, pathlib.Path, pathlib.Path, str]


class TrajectoryDataset(torch.utils.data.Dataset):
    def __init__(
        self,
        train: bool,
        num_images: int,
        dataset_dir: str,
        map_type: str,
        val_ratio: float = 0,
        filter_collision_maps: bool = True,
    ):
        self.dataset_dir = pathlib.Path(dataset_dir)
        self.map_type = map_type

        with open(self.dataset_dir / "meta.yaml") as f:
            self.cfg = yaml.safe_load(f)

        self.collision_maps = set()
        collision_fname = self.dataset_dir / "collision_maps.txt"
        if collision_fname.is_file() and filter_collision_maps:
            with open(collision_fname, "r") as f:
                self.collision_maps.update(f.read().splitlines())

        files_per_type = self.get_all_files()
        all_train_files: List[FileInfo] = []
        all_val_files: List[FileInfo] = []
        num_train = int((1 - val_ratio) * self.cfg["num_envs"])
        for type_ in files_per_type:
            if map_type == "mixed" or map_type == type_:
                all_train_files.extend(files_per_type[type_][:num_train])
                all_val_files.extend(files_per_type[type_][num_train:])

        random.shuffle(all_train_files)
        random.shuffle(all_val_files)
        self.files = (
            all_train_files[:num_images] if train else all_val_files[:num_images]
        )

    def get_all_files(self) -> Dict[str, List[FileInfo]]:
        files: Dict[str, List[FileInfo]] = dict((k, []) for k in self.cfg["map_types"])
        for map_type in self.cfg["map_types"]:
            for idx in range(self.cfg["num_envs"]):
                if f"{map_type}_{idx}" in self.collision_maps:
                    continue
                img_fname = self.dataset_dir / "im_sdf" / map_type / f"{idx}_im.png"
                sdf_fname = self.dataset_dir / "im_sdf" / map_type / f"{idx}_sdf.npy"
                traj_fname = (
                    self.dataset_dir
                    / "opt_trajs_gpmp2"
                    / map_type
                    / f"env_{idx}_prob_0.npz"
                )
                for f in [img_fname, sdf_fname, traj_fname]:
                    assert os.path.isfile(f)
                files[map_type].append(
                    (img_fname, sdf_fname, traj_fname, f"{map_type}_{idx}")
                )
        return files

    def __getitem__(self, idx: int):
        img_file, sdf_file, traj_file, file_id = self.files[idx]

        env_params = self.cfg["env_params"]

        # SDF Data
        cells_per_unit = self.cfg["im_size"] / (
            env_params["x_lims"][1] - env_params["x_lims"][0]
        )
        cell_size = torch.tensor(cells_per_unit).reciprocal().view(1)
        origin = torch.tensor([env_params["x_lims"][0], env_params["y_lims"][0]])
        sdf_data = torch.from_numpy(np.load(sdf_file))

        # Image
        tmp_map = plt.imread(img_file)
        if tmp_map.ndim == 3:
            the_map = tmp_map[..., 0]

        # Trajectory
        traj_data = np.load(traj_file)
        trajectory = torch.from_numpy(traj_data["th_opt"]).permute(1, 0)
        # next two lines re-orient the dgpmp2 trajectory to theseus coordinate system
        trajectory[1] *= -1.0
        trajectory[3] *= -1.0
        return {
            "map_tensor": the_map,
            "sdf_origin": origin.double(),
            "cell_size": cell_size.double(),
            "sdf_data": sdf_data.double(),
            "expert_trajectory": trajectory.double(),
            "file_id": file_id,
        }

    def __len__(self):
        return len(self.files)


# ----------------------------------------------------------------------------------- #
# ------------------------------- Plotting utilities -------------------------------- #
# ----------------------------------------------------------------------------------- #
def _create_line_from_trajectory(
    x_list, y_list, linestyle="-", linewidth=2, alpha=1.0, color="red"
):
    line = plt.Line2D(x_list, y_list)
    line.set_linestyle(linestyle)
    line.set_linewidth(linewidth)
    line.set_color(color)
    line.set_alpha(alpha)
    return line


def _add_robot_to_trajectory(x_list, y_list, radius, color="magenta", alpha=0.05):
    patches = []
    for i in range(x_list.shape[0]):
        patch = mpl.patches.Circle((x_list[i], y_list[i]), radius)
        patches.append(patch)
    patch_collection = mpl.collections.PatchCollection(
        patches, alpha=alpha, color=color
    )
    return patch_collection


def generate_trajectory_figs(
    map_tensor: torch.Tensor,
    sdf: th.eb.SignedDistanceField2D,
    trajectories: List[torch.Tensor],
    robot_radius: float,
    max_num_figures: int = 20,
    labels: Optional[List[str]] = None,
    fig_idx_robot: int = 1,
    figsize: Tuple[int, int] = (8, 8),
) -> List[plt.Figure]:
    # cell rows/cols for each batch of trajectories
    traj_rows = []
    traj_cols = []
    for trajectory in trajectories:
        row, col, _ = sdf.convert_points_to_cell(trajectory[:, :2, :])
        traj_rows.append(np.clip(row, 0, map_tensor.shape[1] - 1))
        traj_cols.append(np.clip(col, 0, map_tensor.shape[1] - 1))

    # Generate a separate figure for each batch index
    colors = ["green", "blue", "red"]
    if not labels:
        labels = ["initial_solution", "best_solution", "expert"]
    figures: List[plt.Figure] = []
    for map_idx in range(map_tensor.shape[0]):
        if map_idx >= max_num_figures:
            continue
        fig, axs = plt.subplots(1, 1, figsize=figsize)
        map_data = map_tensor[map_idx].clone().cpu().numpy()
        map_data = np.tile(map_data, (3, 1, 1)).transpose((1, 2, 0))
        axs.imshow(map_data)
        cell_size = sdf.cell_size.tensor
        patches = []
        for t_idx, trajectory in enumerate(trajectories):
            row = traj_rows[t_idx][map_idx]
            col = traj_cols[t_idx][map_idx]
            line = _create_line_from_trajectory(col, row, color=colors[t_idx])
            axs.add_line(line)
            if t_idx == fig_idx_robot:  # solution trajectory
                radius = robot_radius / cell_size[map_idx][0]
                patch_coll = _add_robot_to_trajectory(col, row, radius)
                axs.add_collection(patch_coll)
            patches.append(mpl.patches.Patch(color=colors[t_idx], label=labels[t_idx]))
        patches.append(
            mpl.patches.Patch(color="magenta", label=f"radius = {robot_radius}")
        )
        axs.legend(handles=patches, fontsize=10)
        fig.tight_layout()
        figures.append(fig)
    return figures



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc

import torch

_LOSS_EPS = 1e-20


class RobustLoss(abc.ABC):
    @classmethod
    def evaluate(cls, x: torch.Tensor, log_radius: torch.Tensor) -> torch.Tensor:
        return cls._evaluate_impl(x, log_radius.exp())

    @classmethod
    def linearize(cls, x: torch.Tensor, log_radius: torch.Tensor) -> torch.Tensor:
        return cls._linearize_impl(x, log_radius.exp())

    @staticmethod
    @abc.abstractmethod
    def _evaluate_impl(x: torch.Tensor, radius: torch.Tensor) -> torch.Tensor:
        pass

    @staticmethod
    @abc.abstractmethod
    def _linearize_impl(x: torch.Tensor, raidus: torch.Tensor) -> torch.Tensor:
        pass


class WelschLoss(RobustLoss):
    @staticmethod
    def _evaluate_impl(x: torch.Tensor, radius: torch.Tensor) -> torch.Tensor:
        return radius - radius * torch.exp(-x / (radius + _LOSS_EPS))

    @staticmethod
    def _linearize_impl(x: torch.Tensor, radius: torch.Tensor) -> torch.Tensor:
        return torch.exp(-x / (radius + _LOSS_EPS))


class HuberLoss(RobustLoss):
    @staticmethod
    def _evaluate_impl(x: torch.Tensor, radius: torch.Tensor) -> torch.Tensor:
        return torch.where(
            x > radius, 2 * torch.sqrt(radius * x.max(radius) + _LOSS_EPS) - radius, x
        )

    @staticmethod
    def _linearize_impl(x: torch.Tensor, radius: torch.Tensor) -> torch.Tensor:
        return torch.sqrt(radius / torch.max(x, radius) + _LOSS_EPS)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th
from theseus.constants import EPS
from theseus.core.tests.common import check_copy_var
from theseus.utils import numeric_jacobian

from .common import (
    check_adjoint,
    check_compose,
    check_exp_map,
    check_inverse,
    check_jacobian_for_local,
    check_log_map,
    check_normalize,
    check_projection_for_compose,
    check_projection_for_exp_map,
    check_projection_for_inverse,
    check_projection_for_log_map,
    check_projection_for_rotate_and_transform,
)


def test_exp_map():
    for batch_size in [1, 20, 100]:
        theta = torch.from_numpy(np.linspace(-np.pi, np.pi, batch_size)).view(-1, 1)
        check_exp_map(theta, th.SO2, EPS)
        check_projection_for_exp_map(theta, th.SO2)


def test_log_map():
    for batch_size in [1, 2, 100]:
        theta = torch.from_numpy(np.linspace(-np.pi, np.pi, batch_size)).view(-1, 1)
        check_log_map(theta, th.SO2, EPS)
        check_projection_for_log_map(theta, th.SO2)


def test_compose():
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        so2_1 = th.SO2.rand(batch_size, generator=rng, dtype=torch.float64)
        so2_2 = th.SO2.rand(batch_size, generator=rng, dtype=torch.float64)
        check_compose(so2_1, so2_2)


def test_inverse():
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        so2 = th.SO2.rand(batch_size, generator=rng, dtype=torch.float64)
        check_inverse(so2)


def test_rotate_and_unrotate():
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size_so2 in [1, 20, 100]:
            for batch_size_pnt in [1, 20, 100]:
                if (
                    batch_size_so2 != 1
                    and batch_size_pnt != 1
                    and batch_size_pnt != batch_size_so2
                ):
                    continue

                so2 = th.SO2.rand(batch_size_so2, generator=rng, dtype=torch.float64)
                # Tests that rotate works from tensor. unrotate() would work similarly), but
                # it's also tested indirectly by test_transform_to() for SE2
                point_tensor = torch.randn(batch_size_pnt, 2).double()
                jacobians_rotate = []
                rotated_point = so2.rotate(point_tensor, jacobians=jacobians_rotate)
                expected_rotated_data = so2.to_matrix() @ point_tensor.unsqueeze(2)
                jacobians_unrotate = []
                unrotated_point = so2.unrotate(rotated_point, jacobians_unrotate)

                # Check the operation result
                assert torch.allclose(
                    expected_rotated_data.squeeze(2), rotated_point.tensor, atol=EPS
                )
                assert torch.allclose(point_tensor, unrotated_point.tensor, atol=EPS)

                # Check the jacobians
                # function_dim = 2 because rotate(theta, (x, y)) --> (x_new, y_new)
                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].rotate(groups[1]),
                    [so2, th.Point2(point_tensor)],
                    function_dim=2,
                )

                assert jacobians_rotate[0].shape == expected_jac[0].shape
                assert jacobians_rotate[1].shape == expected_jac[1].shape
                assert torch.allclose(jacobians_rotate[0], expected_jac[0])
                assert torch.allclose(jacobians_rotate[1], expected_jac[1])

                expected_jac = numeric_jacobian(
                    lambda groups: groups[0].unrotate(groups[1]),
                    [so2, rotated_point],
                    function_dim=2,
                )
                assert jacobians_unrotate[0].shape == expected_jac[0].shape
                assert jacobians_unrotate[1].shape == expected_jac[1].shape
                assert torch.allclose(jacobians_unrotate[0], expected_jac[0])
                assert torch.allclose(jacobians_unrotate[1], expected_jac[1])


def test_adjoint():
    rng = torch.Generator()
    rng.manual_seed(0)
    for batch_size in [1, 20, 100]:
        so2 = th.SO2.rand(batch_size, generator=rng, dtype=torch.float64)
        tangent = torch.randn(batch_size, 1).double()
        check_adjoint(so2, tangent)


def test_copy():
    rng = torch.Generator()
    rng.manual_seed(0)
    so2 = th.SO2.rand(1, generator=rng, dtype=torch.float64)
    check_copy_var(so2)


def test_projection():
    rng = torch.Generator()
    rng.manual_seed(0)
    for _ in range(10):  # repeat a few times
        for batch_size in [1, 20, 100]:
            # Test SO2.rotate
            check_projection_for_rotate_and_transform(
                th.SO2, th.Point2, th.SO2.rotate, batch_size, rng
            )

            # Test SO2.unrotate
            check_projection_for_rotate_and_transform(
                th.SO2, th.Point2, th.SO2.unrotate, batch_size, rng
            )

            # Test SO2.compose
            check_projection_for_compose(th.SO2, batch_size, rng)

            # Test SO2.inverse
            check_projection_for_inverse(th.SO2, batch_size, rng)


def test_local_map():
    rng = torch.Generator()
    rng.manual_seed(0)

    for batch_size in [1, 20, 100]:
        group0 = th.SO2.rand(batch_size)
        group1 = th.SO2.rand(batch_size)

        check_jacobian_for_local(group0, group1, Group=th.SO2, is_projected=True)


@pytest.mark.parametrize("batch_size", [1, 20, 100])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_normalization(batch_size, dtype):
    check_normalize(th.SO2, batch_size, dtype)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math

import torch

TEST_EPS = 5e-7
EPS = 1e-10
PI = math.pi


def _CHECK_DTYPE_SUPPORTED(dtype):
    if dtype not in [torch.float32, torch.float64]:
        raise ValueError(
            f"Unsupported data type {dtype}. "
            "Theseus only supports 32- and 64-bit tensors."
        )


class EPSDict:
    def __init__(self, float32_eps, float64_eps):
        self.float32_eps = float32_eps
        self.float64_eps = float64_eps

    def __getitem__(self, dtype):
        _CHECK_DTYPE_SUPPORTED(dtype)
        if dtype is torch.float32:
            return self.float32_eps
        else:
            return self.float64_eps


_SO2_NORMALIZATION_EPS = EPSDict(float32_eps=1e-12, float64_eps=1e-12)

_SO2_MATRIX_EPS = EPSDict(float32_eps=1e-5, float64_eps=4e-7)

_SE2_NEAR_ZERO_EPS = EPSDict(float32_eps=3e-2, float64_eps=1e-6)

_SO3_NEAR_PI_EPS = EPSDict(float32_eps=1e-2, float64_eps=1e-7)

_SO3_NEAR_ZERO_EPS = EPSDict(float32_eps=1e-2, float64_eps=5e-3)

_SO3_MATRIX_EPS = EPSDict(float32_eps=4e-4, float64_eps=1e-6)

_SO3_QUATERNION_EPS = EPSDict(float32_eps=2e-4, float64_eps=5e-7)

_SO3_HAT_EPS = EPSDict(float32_eps=5e-6, float64_eps=5e-7)

_SE3_NEAR_PI_EPS = EPSDict(float32_eps=1e-2, float64_eps=1e-7)

_SE3_NEAR_ZERO_EPS = EPSDict(float32_eps=1e-2, float64_eps=5e-3)

_SE3_HAT_EPS = EPSDict(float32_eps=5e-6, float64_eps=5e-7)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
from enum import Enum
from typing import Any, Dict, Optional, Type

import numpy as np
import torch

from theseus.core import Objective
from theseus.optimizer import Linearization, Optimizer, OptimizerInfo

from .linear_solver import LinearSolver


class LinearOptimizerStatus(Enum):
    START = 0
    CONVERGED = 1
    FAIL = -1


class LinearOptimizer(Optimizer):
    def __init__(
        self,
        objective: Objective,
        linear_solver_cls: Type[LinearSolver],
        *args,
        vectorize: bool = True,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        linear_solver_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        super().__init__(objective, vectorize=vectorize, **kwargs)
        linearization_kwargs = linearization_kwargs or {}
        linear_solver_kwargs = linear_solver_kwargs or {}
        self.linear_solver = linear_solver_cls(
            objective,
            linearization_cls=linearization_cls,
            linearization_kwargs=linearization_kwargs,
            **linear_solver_kwargs,
        )

    def _optimize_impl(
        self,
        **kwargs,
    ) -> OptimizerInfo:

        info = OptimizerInfo(
            best_solution={},
            status=np.array([LinearOptimizerStatus.START] * self.objective.batch_size),
        )
        try:
            self.linear_solver.linearization.linearize()
            delta = self.linear_solver.solve()
        except RuntimeError as run_err:
            msg = (
                f"There was an error while running the linear optimizer. "
                f"Original error message: {run_err}."
            )
            if torch.is_grad_enabled():
                raise RuntimeError(
                    msg + " Backward pass will not work. To obtain "
                    "the best solution seen before the error, run with torch.no_grad()"
                )
            else:
                warnings.warn(msg, RuntimeWarning)
                info.status[:] = LinearOptimizerStatus.FAIL
                return info
        self.objective.retract_optim_vars(
            delta, self.linear_solver.linearization.ordering
        )
        info.status[:] = LinearOptimizerStatus.CONVERGED
        for var in self.linear_solver.linearization.ordering:
            info.best_solution[var.name] = var.tensor.clone().cpu()
        return info



# Copyright 2004-present Facebook. All Rights Reserved.
#
# Author: Daniel DeTone (ddetone)

import math
import warnings
from typing import NamedTuple, Optional

import torch


def truncated_normal_(N: int, mean: float = 0.0, std: float = 1):
    """Draws N samples from a truncated normal distribution ~N_t(mean,std).

    :param N: number of samples to draw
    :param mean: mean of distribution
    :param std: standard deviation of distribution

    """
    tensor = torch.FloatTensor(N)
    tmp = tensor.new_empty(tensor.shape + (4,)).normal_()
    valid = (tmp < 2) & (tmp > -2)
    ind = valid.max(-1, keepdim=True)[1]
    tensor.data.copy_(tmp.gather(-1, ind).squeeze(-1))
    tensor.data.mul_(std).add_(mean)
    return tensor


def perspective_grid(
    coeffs: torch.Tensor,
    size: torch.Size,
    align_corners: bool = False,
    eps: float = 1e-9,
):
    """
    Generate a grid of u,v sample locations from a perspective transform.
    Acts similarly to torch's affine_grid, but allows for 3x3 perspective warps.

    :param coeffs: Bx8 input batch of perspective transform matrices
    (flattened first 8 elements of 3x3 matrix, assuming last element = 1)
    :param size: NxCxHxW the target output size
    :param align_corners: if True, consider -1 and 1 to refer to the centers
    of the corner pixels
    :param eps: factor to prevent division by zero

    """
    B, H, W = int(size[0]), int(size[-2]), int(size[-1])
    base_grid = torch.empty(B, H, W, 3).to(coeffs)
    # Generate base_grid [-1,1], taking into account align_corners.
    step_x = torch.linspace(-1, 1, W)
    step_y = torch.linspace(-1, 1, H)
    if align_corners is False:
        step_x = step_x * (W - 1) / W
        step_y = step_y * (H - 1) / H
    base_grid[..., 0].copy_(step_x)
    base_grid[..., 1].copy_(step_y.unsqueeze(-1))
    base_grid[..., 2].fill_(1)
    base_grid = base_grid.reshape(B, -1, 3)
    # Generate homography matrix.
    h_mat = torch.ones(B, 9).to(coeffs)
    h_mat[:, :-1] = coeffs
    h_mat = h_mat.reshape(B, 3, 3).transpose(1, 2)
    # Apply homography warp, then normalize by homogenous coordinate.
    grid = base_grid @ h_mat
    grid[:, :, 0] = grid[:, :, 0] / (grid[:, :, 2] + eps)
    grid[:, :, 1] = grid[:, :, 1] / (grid[:, :, 2] + eps)
    grid = grid[:, :, :2].reshape(B, H, W, 2)
    return grid


def check_input(inp: torch.Tensor):
    """Makes sure that input tensor meets input specifications."""
    if inp.dim() != 4:
        raise ValueError("Input tensor must have 4 dims (B, C, H, W)")
    if inp.shape[1] != 1 and inp.shape[1] != 3:
        raise ValueError("Only 1 and 3 channel inputs are supported")
    if inp.dtype != torch.float32:
        raise TypeError("Only torch.float32 supported.")
    if inp.min() < 0.0 or inp.max() > 1.0:
        raise ValueError("Image must be normalized between [0,1].")
    return


# Note(dd): It was a pain to update kornia, so I took this from the kornia directly.
def get_perspective_transform(src, dst):
    r"""Calculates a perspective transform from four pairs of the corresponding
    points.
    The function calculates the matrix of a perspective transform so that:
    .. math ::
        \begin{bmatrix}
        t_{i}x_{i}^{'} \\
        t_{i}y_{i}^{'} \\
        t_{i} \\
        \end{bmatrix}
        =
        \textbf{map_matrix} \cdot
        \begin{bmatrix}
        x_{i} \\
        y_{i} \\
        1 \\
        \end{bmatrix}
    where
    .. math ::
        dst(i) = (x_{i}^{'},y_{i}^{'}), src(i) = (x_{i}, y_{i}), i = 0,1,2,3
    Args:
        src (torch.Tensor): coordinates of quadrangle vertices in the source image
        with shape :math:`(B, 4, 2)`.
        dst (torch.Tensor): coordinates of the corresponding quadrangle vertices in
            the destination image with shape :math:`(B, 4, 2)`.
    Returns:
        torch.Tensor: the perspective transformation with shape :math:`(B, 3, 3)`.
    .. note::
        This function is often used in conjuntion with :func:`warp_perspective`.
    """

    # we build matrix A by using only 4 point correspondence. The linear
    # system is solved with the least square method, so here
    # we could even pass more correspondence
    p = []
    for i in [0, 1, 2, 3]:
        p.append(_build_perspective_param(src[:, i], dst[:, i], "x"))
        p.append(_build_perspective_param(src[:, i], dst[:, i], "y"))

    # A is Bx8x8
    A = torch.stack(p, dim=1)

    # b is a Bx8x1
    b = torch.stack(
        [
            dst[:, 0:1, 0],
            dst[:, 0:1, 1],
            dst[:, 1:2, 0],
            dst[:, 1:2, 1],
            dst[:, 2:3, 0],
            dst[:, 2:3, 1],
            dst[:, 3:4, 0],
            dst[:, 3:4, 1],
        ],
        dim=1,
    )

    # solve the system Ax = b
    X = torch.linalg.solve(A, b)

    # create variable to return
    batch_size = src.shape[0]
    M = torch.ones(batch_size, 9, device=src.device, dtype=src.dtype)
    M[..., :8] = torch.squeeze(X, dim=-1)

    return M.view(-1, 3, 3)  # Bx3x3


def _build_perspective_param(
    p: torch.Tensor, q: torch.Tensor, axis: str
) -> torch.Tensor:
    ones = torch.ones_like(p)[..., 0:1]
    zeros = torch.zeros_like(p)[..., 0:1]
    if axis == "x":
        return torch.cat(
            [
                p[:, 0:1],
                p[:, 1:2],
                ones,
                zeros,
                zeros,
                zeros,
                -p[:, 0:1] * q[:, 0:1],
                -p[:, 1:2] * q[:, 0:1],
            ],
            dim=1,
        )

    elif axis == "y":
        return torch.cat(
            [
                zeros,
                zeros,
                zeros,
                p[:, 0:1],
                p[:, 1:2],
                ones,
                -p[:, 0:1] * q[:, 1:2],
                -p[:, 1:2] * q[:, 1:2],
            ],
            dim=1,
        )
    else:
        raise ValueError("Bad input axis, should be x or y")


class GeoAugParam(NamedTuple):
    min: Optional[float] = 0.0
    max: Optional[float] = 0.0


class RandomGeoAug(object):
    """
    Applies batch-wise geometric augmentations to images.
    Requires that input/output tensors are sized BxCxHxW.
    """

    def __init__(
        self,
        rotate_param: Optional[GeoAugParam] = None,
        scale_param: Optional[GeoAugParam] = None,
        translate_x_param: Optional[GeoAugParam] = None,
        translate_y_param: Optional[GeoAugParam] = None,
        shear_x_param: Optional[GeoAugParam] = None,
        shear_y_param: Optional[GeoAugParam] = None,
        perspective_param: Optional[GeoAugParam] = None,
    ):

        self.rotate_param = (
            GeoAugParam(min=-180, max=180) if rotate_param is None else rotate_param
        )
        self.scale_param = (
            GeoAugParam(min=0.5, max=2.0) if scale_param is None else scale_param
        )
        self.translate_x_param = (
            GeoAugParam(min=-0.3, max=0.3)
            if translate_x_param is None
            else translate_x_param
        )
        self.translate_y_param = (
            GeoAugParam(min=-0.3, max=0.3)
            if translate_y_param is None
            else translate_y_param
        )
        self.shear_x_param = (
            GeoAugParam(min=-20, max=20) if shear_x_param is None else shear_x_param
        )
        self.shear_y_param = (
            GeoAugParam(min=-20, max=20) if shear_y_param is None else shear_y_param
        )
        self.perspective_param = (
            GeoAugParam(min=-0.3, max=0.3)
            if perspective_param is None
            else perspective_param
        )
        self.eps = 1e-9

        self.param_names = []
        for key in dir(self):
            if "_param" in key:
                self.param_names.append(key)

    def __repr__(self):
        ret = self.__class__.__name__ + "(\n"
        for key in dir(self):
            if "_param" in key:
                ret += ("  " + key + " = " + getattr(self, key).__repr__()) + ",\n"
        ret += ")"
        return ret

    def _get_perspective_matrix(
        self,
        angle: torch.Tensor,
        translate: torch.Tensor,
        scale: torch.Tensor,
        shear: torch.Tensor,
        delta_corners: torch.Tensor,
    ):
        B = angle.shape[0]
        # Thus, the inverse is M^-1 = RSS^-1 * T^-1
        rot = torch.deg2rad(-angle)
        sx, sy = [torch.deg2rad(s) for s in shear.t()]
        tx, ty = 2 * translate[:, 0], 2 * translate[:, 1]
        # RSS without scaling
        a = torch.cos(rot - sy) / torch.cos(sy)
        b = -torch.cos(rot - sy) * torch.tan(sx) / torch.cos(sy) - torch.sin(rot)
        c = torch.sin(rot - sy) / torch.cos(sy)
        d = -torch.sin(rot - sy) * torch.tan(sx) / torch.cos(sy) + torch.cos(rot)
        # Inverted rotation matrix with scale and shear
        zeros = torch.zeros_like(a)
        matrix = torch.stack([d, -b, zeros, -c, a, zeros], dim=-1)
        matrix = torch.stack([x / scale for x in matrix.t()], dim=-1)
        # Apply inverse of translation and of center translation: RSS^-1 * T^-1
        matrix[:, 2] += matrix[:, 0] * (-tx) + matrix[:, 1] * (-ty)
        matrix[:, 5] += matrix[:, 3] * (-tx) + matrix[:, 4] * (-ty)
        matrix = torch.cat([matrix, matrix.new_zeros((B, 3))], dim=1)
        matrix[:, -1] = 1.0
        matrix = matrix.reshape(B, 3, 3)
        # Generate non-affine effect by perturbing the four corners.
        start = torch.ones(B, 4, 2).to(a)
        end = start - delta_corners  # Positive delta takes corners inwards.
        corner_signs = torch.tensor([[-1, -1], [1, -1], [1, 1], [-1, 1]]).to(a)
        corner_signs = corner_signs.repeat(B, 1, 1)
        start *= corner_signs
        end *= corner_signs
        # Set up a linear system to solve for the homography.
        # pmat = solve_homography(start, end)
        # pmat = solve_homography_old(start, end)
        # pmat = kornia.homography.find_homography_dlt(start, end) # Newer kornia version API.
        # pmat = kornia.get_perspective_transform(start, end)
        pmat = get_perspective_transform(
            start, end
        )  # Copied kornia functioned directly.
        # Apply perspective transform first, then affine.
        matrix = pmat @ matrix
        # Normalize such that H[2,2] = 1.
        eps = 1e-9
        matrix /= matrix[:, 2, 2].clone().reshape(B, 1, 1) + eps
        return matrix

    def _generate_transform(self, B: int, device: torch.device):
        rot_range = (self.rotate_param.min, self.rotate_param.max)
        trans_x_range = (self.translate_x_param.min, self.translate_x_param.max)
        trans_y_range = (self.translate_y_param.min, self.translate_y_param.max)
        shear_x_range = (self.shear_x_param.min, self.shear_x_param.max)
        shear_y_range = (self.shear_y_param.min, self.shear_y_param.max)
        scale_range = (self.scale_param.min, self.scale_param.max)
        pscale_range = (self.perspective_param.min, self.perspective_param.max)
        # Randomly generate parameters that define the perspective transform.
        rot = torch.empty(B, device=device).uniform_(rot_range[0], rot_range[1])
        trans_x = torch.empty(B, device=device).uniform_(
            trans_x_range[0], trans_x_range[1]
        )
        trans_y = torch.empty(B, device=device).uniform_(
            trans_y_range[0], trans_y_range[1]
        )
        trans_xy = torch.stack([trans_x, trans_y], dim=-1)
        shear_x = torch.empty(B, device=device).uniform_(
            shear_x_range[0], shear_x_range[1]
        )
        shear_y = torch.empty(B, device=device).uniform_(
            shear_y_range[0], shear_y_range[1]
        )
        shear_xy = torch.stack([shear_x, shear_y], dim=-1)
        scale = torch.empty(B, device=device).uniform_(scale_range[0], scale_range[1])
        delta_corners = torch.empty(B, 4, 2, device=device).uniform_(
            pscale_range[0], pscale_range[1]
        )
        # Given parameters, generate 3x3 perspective matrices.
        transform = self._get_perspective_matrix(
            rot, trans_xy, scale, shear_xy, delta_corners
        )
        return transform

    def _apply_perspective(
        self, img: torch.Tensor, transform: torch.Tensor, align_corners: bool = True
    ):
        coeffs = transform.reshape(-1, 9)[:, :-1]
        grid2 = perspective_grid(
            coeffs, img.shape, align_corners=align_corners, eps=self.eps
        )
        img2 = torch.nn.functional.grid_sample(img, grid2, align_corners=align_corners)
        return img2

    def _denormalize_transform(self, transform: torch.Tensor, H: int, W: int):
        B = int(transform.shape[0])
        ones = transform.new_ones((B, 1))
        zeros = transform.new_zeros((B, 1))
        scaleW = (1.0 / ((W - 1.0) / 2.0)) * ones
        scaleH = (1.0 / ((H - 1.0) / 2.0)) * ones
        T = torch.stack(
            [scaleW, zeros, -ones, zeros, scaleH, -ones, zeros, zeros, ones], dim=-1
        ).reshape(-1, 3, 3)
        transform2 = torch.inverse(T) @ transform @ T
        transform2 /= transform2[:, 2, 2].reshape(B, 1, 1).clone() + self.eps
        return transform2

    def set_all_identity(self):
        for name in self.param_names:
            if "scale_param" in name:
                val = 1.0
            else:
                val = 0.0
            setattr(self, name, GeoAugParam(min=val, max=val))

    def forward(self, inp, return_transform=False, normalize_returned_transform=False):
        """
        Runs the Random Geometric Augmentation on a batch of images.

        :param inp: BxCxHxW batch of input images
        :param return_transform: if True, also return the transform used to
        warp the inp (in unnormalize image coordinates by default)
        :param normalize_returned_transform: if True, the returned transform
        will be in normalized pixel coords (e.g. [-1,-1] is top left pixel,
        [1,1] is bottom right)

        """
        check_input(inp)  # Make sure input is well-formed.
        if self.scale_param.min <= 0.0 or self.scale_param.max <= 0.0:
            warnings.warn(
                "Scale min and max are are too extreme, resulting image might "
                "be black, range should be (0, inf)",
                RuntimeWarning,
            )
        if self.translate_x_param.min < -1.0 or self.translate_x_param.max > 1.0:
            warnings.warn(
                "Translate_x min and max are too extreme, resulting image might"
                "be black, range should be [-1.0, 1.0]",
                RuntimeWarning,
            )
        if self.translate_y_param.min < -1.0 or self.translate_y_param.max > 1.0:
            warnings.warn(
                "Translate_y min and max are too extreme, resulting image "
                "might be black, range should be [-1.0, 1.0]",
                RuntimeWarning,
            )
        if self.shear_x_param.min < -90 or self.shear_x_param.max > 90:
            warnings.warn(
                "Shear_x min and max are too extreme, resulting image might "
                "be black, range should be [-90, 90]",
                RuntimeWarning,
            )
        if self.shear_y_param.min < -90 or self.shear_y_param.max > 90:
            warnings.warn(
                "Shear_x min and max are too extreme, resulting image might "
                "be black, range should be [-90, 90]",
                RuntimeWarning,
            )
        if self.perspective_param.min < -1.0 or self.perspective_param.max < -1.0:
            warnings.warn(
                "Perspective min and max are too extreme, resulting image "
                "might be black, range should be [-1, inf]",
                RuntimeWarning,
            )

        # Forcing this to be generated on CPU because it's faster.
        inv_transform = self._generate_transform(
            inp.shape[0], device=torch.device("cpu")
        )
        inv_transform = inv_transform.to(inp)
        identity = (
            torch.eye(3).unsqueeze(0).repeat(inp.shape[0], 1, 1).to(inv_transform)
        )
        if torch.allclose(identity, inv_transform):  # Shortcut operation.
            inp2 = inp
            transform = inv_transform
        else:
            inp2 = self._apply_perspective(inp, inv_transform, align_corners=True)
            transform = torch.inverse(inv_transform)
            transform /= transform[:, 2, 2].reshape(-1, 1, 1).clone() + self.eps
        if not return_transform:
            return inp2
        else:
            if normalize_returned_transform:
                return inp2, transform
            else:
                H, W = inp.shape[-2], inp.shape[-1]
                denorm_transform = self._denormalize_transform(transform, H, W)
                return inp2, denorm_transform


class PhotoAugParam(NamedTuple):
    prob: Optional[float] = 0.2
    mag: Optional[float] = 0.5
    fix_pos: Optional[bool] = False


class RandomPhotoAug(object):
    """
    Applies batch-wise photometric distortions to images.
    Requires that input/output tensors are sized BxCxHxW.

    """

    def __init__(
        self,
        contrast_param: Optional[PhotoAugParam] = None,
        sharpen_param: Optional[PhotoAugParam] = None,
        exposure_param: Optional[PhotoAugParam] = None,
        gamma_param: Optional[PhotoAugParam] = None,
        gaussian_smooth_param: Optional[PhotoAugParam] = None,
        motion_blur_param: Optional[PhotoAugParam] = None,
        shadow_highlight_param: Optional[PhotoAugParam] = None,
        gaussian_noise_param: Optional[PhotoAugParam] = None,
        salt_and_pepper_param: Optional[PhotoAugParam] = None,
    ):

        self.contrast_param = (
            PhotoAugParam() if contrast_param is None else contrast_param
        )
        self.sharpen_param = PhotoAugParam() if sharpen_param is None else sharpen_param
        self.exposure_param = (
            PhotoAugParam() if exposure_param is None else exposure_param
        )
        self.gamma_param = PhotoAugParam() if gamma_param is None else gamma_param
        self.gaussian_smooth_param = (
            PhotoAugParam() if gaussian_smooth_param is None else gaussian_smooth_param
        )
        self.motion_blur_param = (
            PhotoAugParam() if motion_blur_param is None else motion_blur_param
        )
        self.shadow_highlight_param = (
            PhotoAugParam()
            if shadow_highlight_param is None
            else shadow_highlight_param
        )
        # Fix the order in which the last two are applied by default.
        self.gaussian_noise_param = (
            PhotoAugParam(fix_pos=True)
            if gaussian_noise_param is None
            else gaussian_noise_param
        )
        self.salt_and_pepper_param = (
            PhotoAugParam(fix_pos=True)
            if salt_and_pepper_param is None
            else salt_and_pepper_param
        )

        # Sets the default order.
        self.fn_names = [
            "contrast",
            "sharpen",
            "exposure",
            "gamma",
            "gaussian_smooth",
            "motion_blur",
            "shadow_highlight",
            "gaussian_noise",
            "salt_and_pepper",
        ]

        # Each fn_name must be a method.
        for name in self.fn_names:
            assert hasattr(self, name) and callable(getattr(self, name))

        self.param_names = []
        for key in dir(self):
            if "_param" in key:
                self.param_names.append(key)

    def __repr__(self):
        ret = self.__class__.__name__ + "(\n"
        for key in dir(self):
            if key.endswith("_param"):
                ret += ("  " + key + " = " + getattr(self, key).__repr__()) + ",\n"
        ret += ")"
        return ret

    def set_all_probs(self, val: int):
        for name in self.param_names:
            param = getattr(self, name)
            setattr(self, name, param._replace(prob=val))

    def set_all_mags(self, val: int):
        for name in self.param_names:
            param = getattr(self, name)
            setattr(self, name, param._replace(mag=val))

    def set_all_fix_pos(self, val: bool):
        for name in self.param_names:
            param = getattr(self, name)
            setattr(self, name, param._replace(fix_pos=val))

    def contrast(self, inp: torch.Tensor, mag: float = 0.5):
        B = inp.shape[0]
        # Kind of arbitrary, asymmetric normal distribution with
        # has median=1 and doesn't have clipping artifacts.
        scales = truncated_normal_(B, mean=1.0, std=(mag * 0.9 * 0.5 + 1e-9))
        scales = torch.where(scales > 1.0, scales + (scales - 1.0) * 3.0, scales)
        scales = scales.to(inp)
        out = inp - 0.5
        out = out * scales.reshape(-1, 1, 1, 1)
        out = out + 0.5
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def sharpen(self, inp: torch.Tensor, mag: float = 0.5):
        B, C, H, W = inp.shape
        scale = truncated_normal_(B, mean=0, std=mag).abs()
        val = (1.0 / 5.0) + scale
        kernel = torch.zeros(B, 3, 3).to(inp)
        kernel[:, 1, 1] = val * 5
        kernel[:, 0, 1] = -((5 * val - 1.0) / 4.0)
        kernel[:, 1, 0] = -((5 * val - 1.0) / 4.0)
        kernel[:, 1, 2] = -((5 * val - 1.0) / 4.0)
        kernel[:, 2, 1] = -((5 * val - 1.0) / 4.0)
        # Use grouped conv to apply different kernel to each batch element.
        kernel = kernel[:, None, ...]
        kernel = torch.repeat_interleave(kernel, C, dim=0)
        inp = inp.reshape(1, B * C, H, W)
        out = torch.nn.functional.conv2d(inp, kernel, padding=1, groups=B * C)
        out = out.reshape(B, C, H, W)
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def exposure(self, inp: torch.Tensor, mag: float = 0.5):
        B = inp.shape[0]
        max_delta = mag * 0.9  # Don't let max pixel value go below 0.1.
        new_maxs = truncated_normal_(B, mean=1.0, std=0.5 * max_delta).to(inp)
        out = inp * new_maxs.reshape(-1, 1, 1, 1)
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def gamma(self, inp: torch.Tensor, mag: float = 0.5):
        B = inp.shape[0]
        # Kind of arbitrary, asymmetric normal distribution with
        # has median=1 and doesn't have clipping artifacts.
        gam = truncated_normal_(B, mean=1.0, std=mag * 0.9 * 0.5).to(inp)
        gam = torch.where(gam > 1.0, gam + (gam - 1.0) * 3.0, gam)
        out = (inp ** gam.reshape(-1, 1, 1, 1)).clamp(0, 1)
        return out

    def gaussian_smooth(
        self,
        inp: torch.Tensor,
        kmin: int = 1,
        kmax: int = 11,
        smin: float = 1.0,
        smax: float = 5.0,
        mag: float = 0.5,
    ):
        B, C, H, W = inp.shape
        # Maximum possible kernel size.
        kmax = int(kmax * mag) + 1
        if kmax == 1:  # 1x1 gaussian blur does nothing.
            return inp
        half_sizes = torch.randint(size=(B,), low=kmin, high=kmax)  # check kmax
        # To batch the operations, we use the largest kernel size.
        kernel_size = int(half_sizes.max().item() * 2 + 1)
        sigma = mag * torch.FloatTensor(B).uniform_(smin, smax)
        x_cord = torch.arange(kernel_size)
        x_grid = x_cord.repeat(kernel_size).view(kernel_size, kernel_size)
        y_grid = x_grid.t()
        xy_grid = torch.stack([x_grid, y_grid], dim=-1)
        xy_grid = xy_grid.unsqueeze(0).repeat(B, 1, 1, 1)
        mean = (kernel_size - 1) / 2.0
        variance = sigma**2.0
        variance = variance.reshape(-1, 1, 1, 1)
        gaussian_kernel = (1.0 / (2.0 * math.pi * variance)) * torch.exp(
            -torch.sum((xy_grid - mean) ** 2.0, dim=-1, keepdim=True) / (2 * variance)
        )
        # Make sure sum of values in gaussian kernel equals 1.
        kernel = gaussian_kernel / torch.sum(
            gaussian_kernel, dim=[1, 2, 3], keepdim=True
        )
        kernel = kernel.to(inp)
        # Use grouped conv to apply different kernel to each batch element.
        kernel = kernel.reshape(B, 1, kernel_size, kernel_size)
        kernel = torch.repeat_interleave(kernel, C, dim=0)
        kernel = kernel.reshape(B * C, 1, kernel_size, kernel_size)
        inp = inp.reshape(1, B * C, H, W)
        out = torch.nn.functional.conv2d(
            inp, kernel, padding=int(half_sizes.max()), groups=B * C
        )
        out = out.reshape(B, C, H, W)
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def motion_blur(
        self, inp: torch.Tensor, mag: float = 0.5, kmin: int = 1, kmax: int = 11
    ):
        B, C, H, W = inp.shape
        # Maximum possible kernel size.
        kmax = int(kmax * mag) + 1
        if kmax == 1:  # 1x1 gaussian blur does nothing.
            return inp
        half_sizes = torch.randint(size=(B,), low=kmin, high=kmax)  # check kmax
        # To batch the operations, we use the largest kernel size.
        kernel_sizes = half_sizes * 2 + 1
        kernel_size = int(kernel_sizes.max().item())
        # Generate a different motion blur kernel for each batch.
        kernel = torch.zeros((B, kernel_size, kernel_size)).to(inp)
        # TODO(dd): figure out how to vectorize this.
        for i, ks in enumerate(kernel_sizes):
            off = torch.div((kernel_size - ks), 2, rounding_mode="floor")
            start = int(off)
            end = int(kernel_size - off)
            kh = torch.div(kernel_size, 2, rounding_mode="floor")
            kernel[i, kh, start:end] = torch.ones((1, ks)) / ks
        kernel = kernel[:, None, :, :]
        # Rotate the motion blur kernel.
        angle = torch.FloatTensor(B).uniform_(0, 360.0).unsqueeze(0).to(inp)
        rads = (angle * (math.pi / 180.0)).reshape(B, 1, 1)
        r00 = torch.cos(rads)
        r01 = -torch.sin(rads)
        r10 = torch.sin(rads)
        r11 = torch.cos(rads)
        top = torch.cat([r00, r01, 0 * r00], dim=2)
        bot = torch.cat([r10, r11, 0 * r00], dim=2)
        affine = torch.cat([top, bot], dim=1).to(inp)
        grid = torch.nn.functional.affine_grid(
            affine, list(kernel.shape), align_corners=True
        )
        kernel = torch.nn.functional.grid_sample(kernel, grid, align_corners=True)
        # Apply the rotated motion blur kernel to the images.
        kernel = torch.repeat_interleave(kernel, C, dim=0)
        kernel = kernel.reshape(B * C, 1, kernel_size, kernel_size)
        inp = inp.reshape(1, B * C, H, W)
        out = torch.nn.functional.conv2d(
            inp, kernel, padding=int(half_sizes.max()), groups=B * C
        )
        out = out.reshape(B, C, H, W)
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def shadow_highlight(
        self,
        inp: torch.Tensor,
        mag: float = 0.5,
        min_quads: int = 2,
        max_quads: int = 5,
    ):
        B, H, W = inp.shape[0], inp.shape[-2], inp.shape[-1]
        out = inp
        num_quads = torch.randint(low=min_quads, high=max_quads, size=(B,))

        kwargs = {
            "scale_param": GeoAugParam(min=0.5, max=0.9),
            "translate_x_param": GeoAugParam(min=-0.8, max=0.8),
            "translate_y_param": GeoAugParam(min=-0.8, max=0.8),
            "rotate_param": GeoAugParam(min=-180, max=180),
            "perspective_param": GeoAugParam(min=0.0, max=0.8),
        }
        rga = RandomGeoAug(**kwargs)
        max_num_quads = num_quads.max()
        for i in range(max_num_quads):
            # Faster to apply conv to mini image then resize later.
            mask = torch.ones(B, 1, 50, 50).to(inp)
            # Randomly distort the quad.
            mask = rga.forward(mask)
            # Smooth the quad.
            mask = self.gaussian_smooth(mask, kmin=3, kmax=11, smin=3, smax=11, mag=1.0)
            mask = torch.nn.functional.interpolate(
                mask, size=[H, W], mode="bilinear", align_corners=True
            )
            val = truncated_normal_(B, mean=0.0, std=(mag * 0.6)).to(inp)
            quad = mask * val.reshape(B, 1, 1, 1)
            out2 = (1.0 - mask) * out + mask * (out + quad)
            apply_quad = torch.tensor(i).reshape(1, 1, 1, 1).repeat(
                B, 1, 1, 1
            ) < num_quads.reshape(B, 1, 1, 1)
            apply_quad = apply_quad.to(inp.device)
            out = torch.where(apply_quad, out2, out)
            out = torch.clamp(out, 0.0, 1.0)
        return out

    def gaussian_noise(self, inp: torch.Tensor, mag: float = 0.5):
        B, C, H, W = inp.shape
        val = torch.FloatTensor(B).uniform_(0, mag * 0.1).abs().reshape(B, 1, 1)
        noise = val * torch.randn((B, H, W))
        noise = noise[:, None, :, :]
        noise = noise.repeat(1, C, 1, 1).to(inp.device)
        out = inp + noise
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def salt_and_pepper(self, inp: torch.Tensor, mag: float = 0.5):
        B, C, H, W = inp.shape
        noise_prob = mag * torch.FloatTensor(B).normal_(
            0.0, (mag * 0.05 + 1e-9)
        ).abs().to(inp)
        coins = torch.rand(size=(B, H, W)).to(inp).unsqueeze(1).repeat(1, C, 1, 1)
        heavy_noise = 0.5 * torch.randn(B, H, W).to(inp).unsqueeze(1).repeat(1, C, 1, 1)
        out = torch.where(
            coins > noise_prob.reshape(-1, 1, 1, 1), inp, inp + heavy_noise
        )
        out = torch.clamp(out, 0.0, 1.0)
        return out

    def forward(self, inp: torch.Tensor, debug: bool = False, seed: int = None):
        """
        Runs the Random Photometric Augmentation on a batch of images.

        :param inp: BxCxHxW batch of input images
        :param debug: if True, print out some info about which transforms were used

        """
        if seed is not None:
            torch.manual_seed(seed)
        check_input(inp)  # Make sure input is well-formed.
        B, C, H, W = inp.shape
        result = inp.clone()  # Creates a copy of input (leaves input unchanged).
        # Randomly permute function ordering, except if "fix_pos" is True.
        permute_names = [
            name for name in self.fn_names if not getattr(self, name + "_param").fix_pos
        ]
        indices = torch.randperm(len(permute_names)).tolist()
        perm_fns = [permute_names[idx] for idx in indices]
        for i, name in enumerate(self.fn_names):
            if getattr(self, name + "_param").fix_pos:
                perm_fns.insert(i, name)
        # Iterate randomly permuted augmentation functions and maybe apply it.
        for fn_name in perm_fns:
            prob = getattr(self, fn_name + "_param").prob
            coins = torch.rand(B).reshape(-1, 1, 1, 1).to(inp)
            mask = coins < prob
            # Optional printing, helpful for debugging.
            if debug:
                print(fn_name, int(mask.sum()), "/", result.shape[0])
            # Skip operation if no elt in batch needs this augmentation.
            if torch.all(torch.eq(mask, False)):
                continue
            # Run the augmentation across all elts in batch.
            mag = getattr(self, fn_name + "_param").mag
            fn_method = getattr(self, fn_name)
            pre = result.clone()
            # Actual call to augmentation is here.
            result2 = fn_method(pre, mag=mag)
            mask = mask.repeat(1, C, H, W)
            # Apply the augmentation only where needed (from coins).
            final = torch.where(mask, result2, result)
            result = final.clone()
        return result



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .sparse_matrix_utils import random_sparse_binary_matrix
from .utils import build_mlp, gather_from_rows_cols, numeric_jacobian



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest  # noqa: F401

import theseus as th

from .common import run_nonlinear_least_squares_check


def test_gauss_newton():
    run_nonlinear_least_squares_check(th.GaussNewton, {})



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple, Union, cast

import torch

from theseus.geometry.manifold import Manifold

from .lie_group import LieGroup


class Vector(LieGroup):
    def __init__(
        self,
        dof: Optional[int] = None,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: Optional[torch.dtype] = None,
    ):
        if dof is not None and tensor is not None:
            raise ValueError("Please provide only one of dof or tensor.")
        if dof is None:
            if tensor is not None and tensor.ndim == 1:
                tensor = tensor.view(1, -1)
            if tensor is None or tensor.ndim != 2:
                raise ValueError(
                    "If dof not provided, tensor must " "have shape (batch_size, dof)"
                )
            dof = tensor.shape[1]

        super().__init__(dof, tensor=tensor, name=name, dtype=dtype)

    # Vector variables are of shape [batch_size, dof]
    @staticmethod
    def _init_tensor(dof: int) -> torch.Tensor:  # type: ignore
        return torch.zeros(1, dof)

    def dof(self) -> int:
        return self.tensor.shape[1]

    @staticmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "Vector":
        if len(size) != 2:
            raise ValueError("The size should be 2D.")
        return Vector(
            tensor=torch.rand(
                size,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    @staticmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "Vector":
        if len(size) != 2:
            raise ValueError("The size should be 2D.")
        return Vector(
            tensor=torch.randn(
                size,
                generator=generator,
                dtype=dtype,
                device=device,
                requires_grad=requires_grad,
            )
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}"
            f"(dof={self.tensor.shape[1]}, tensor={self.tensor}, name={self.name})"
        )

    def allclose(self, other: "Vector", *args, **kwargs) -> bool:
        return torch.allclose(self.tensor, other.tensor, *args, **kwargs)

    def __add__(self, other: "Vector") -> "Vector":
        return self.__class__._compose_impl(self, other)

    def __sub__(self, other: "Vector") -> "Vector":
        return self.__class__(tensor=torch.sub(self.tensor, other.tensor))

    def __neg__(self) -> "Vector":
        return self.__class__._inverse_impl(self)

    def __mul__(self, other: Union["Vector", torch.Tensor]) -> "Vector":
        if isinstance(other, Vector):
            return self.__class__(tensor=torch.mul(self.tensor, other.tensor))
        elif isinstance(other, torch.Tensor):
            return self.__class__(tensor=torch.mul(self.tensor, other))
        else:
            raise ValueError(
                f"expected type 'Vector' or 'Tensor', got {type(other).__name__}"
            )

    __rmul__ = __mul__

    def __matmul__(self, other: torch.Tensor) -> torch.Tensor:
        if not isinstance(other, torch.Tensor):
            raise ValueError("Vector matmul only accepts torch tensors.")
        if other.ndim != 3:
            raise ValueError(
                f"Vector matmul only accepts tensors with ndim=3 "
                f"but tensor has ndim={other.ndim}."
            )
        return torch.bmm(self.tensor.unsqueeze(1), other).squeeze(1)

    def __rmatmul__(self, other: torch.Tensor) -> torch.Tensor:
        if isinstance(other, Vector):
            raise ValueError("Vector matmul only accepts torch tensors.")
        if other.ndim != 3:
            raise ValueError(
                f"Vector matmul only accepts tensors with ndim=3 but "
                f"tensor has ndim={other.ndim}."
            )
        return torch.bmm(other, self.tensor.unsqueeze(2)).squeeze(2)

    def __truediv__(self, other: Union["Vector", torch.Tensor]) -> "Vector":
        if isinstance(other, Vector):
            return self.__class__(tensor=torch.div(self.tensor, other.tensor))
        elif isinstance(other, torch.Tensor):
            return self.__class__(tensor=torch.div(self.tensor, other))
        else:
            raise ValueError(
                f"expected type 'Vector' or 'Tensor', got {type(other).__name__}"
            )

    def dot(self, other: "Vector") -> torch.Tensor:
        return torch.mul(self.tensor, other.tensor).sum(-1)

    inner = dot

    def abs(self) -> "Vector":
        return self.__class__(tensor=torch.abs(self.tensor))

    def outer(self, other: "Vector") -> torch.Tensor:
        return torch.matmul(self.tensor.unsqueeze(2), other.tensor.unsqueeze(1))

    def norm(self, *args, **kwargs) -> torch.Tensor:
        return torch.norm(self.tensor, *args, **kwargs)

    def cat(self, vecs: Union["Vector", Tuple["Vector"], List["Vector"]]) -> "Vector":
        if isinstance(vecs, Vector):
            result = torch.cat([self.tensor, vecs.tensor], 1)
        else:
            result = torch.cat([self.tensor] + [vec.tensor for vec in vecs], 1)
        return Vector(tensor=result)

    def to_matrix(self) -> torch.Tensor:
        return self.tensor.clone()

    def _local_impl(
        self, vec2: Manifold, jacobians: List[torch.Tensor] = None
    ) -> torch.Tensor:
        if not isinstance(vec2, Vector):
            raise ValueError("Non-vector inputs for Vector.local()")
        else:
            return LieGroup._local_impl(self, vec2, jacobians)

    def _retract_impl(self, delta: torch.Tensor) -> "Vector":
        return self.__class__(tensor=torch.add(self.tensor, delta))

    def _compose_impl(self, vec2: LieGroup) -> "Vector":
        return self.__class__(tensor=torch.add(self.tensor, vec2.tensor))

    def _compose_jacobian(self, _: LieGroup) -> Tuple[torch.Tensor, torch.Tensor]:
        identity = torch.eye(self.dof(), dtype=self.dtype, device=self.device).repeat(
            self.shape[0], 1, 1
        )
        return (identity, identity.clone())

    def _inverse_impl(self) -> "Vector":
        return self.__class__(tensor=-self.tensor)

    def _adjoint_impl(self) -> torch.Tensor:
        return (
            torch.eye(self.dof(), dtype=self.dtype, device=self.device)
            .repeat(self.shape[0], 1, 1)
            .to(self.device)
        )

    def _project_impl(
        self, euclidean_grad: torch.Tensor, is_sparse: bool = False
    ) -> torch.Tensor:
        self._project_check(euclidean_grad, is_sparse)
        return euclidean_grad.clone()

    @staticmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "Vector":
        if tangent_vector.ndim != 2:
            raise ValueError("The dimension of tangent vectors should be 2.")

        Vector._exp_map_jacobian_impl(tangent_vector, jacobians)

        return Vector(tensor=tangent_vector.clone())

    @staticmethod
    def _check_tensor_impl(tensor: torch.Tensor) -> bool:
        if tensor.ndim != 2:
            raise ValueError("Vector variables expect tensors with ndim=2.")

        return True

    @staticmethod
    def normalize(tensor: torch.Tensor) -> torch.Tensor:
        if tensor.ndim != 2:
            raise ValueError("Vector variables expect tensors with ndim=2.")

        return tensor

    @staticmethod
    def _exp_map_jacobian_impl(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]]
    ):
        if jacobians is not None:
            shape = tangent_vector.shape
            LieGroup._check_jacobians_list(jacobians)
            jacobians.append(
                torch.eye(
                    shape[1], dtype=tangent_vector.dtype, device=tangent_vector.device
                ).repeat(shape[0], 1, 1)
            )

    def _log_map_impl(
        self, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        self._log_map_jacobian_impl(jacobians)
        return self.tensor.clone()

    def _log_map_jacobian_impl(self, jacobians: Optional[List[torch.Tensor]] = None):
        if jacobians is not None:
            shape = self.shape
            Vector._check_jacobians_list(jacobians)
            jacobians.append(
                torch.eye(shape[1], dtype=self.dtype, device=self.device).repeat(
                    shape[0], 1, 1
                )
            )

    def __hash__(self):
        return id(self)

    def _copy_impl(self, new_name: Optional[str] = None) -> "Vector":
        return self.__class__(tensor=self.tensor.clone(), name=new_name)

    # added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "Vector":
        return cast(Vector, super().copy(new_name=new_name))


rand_vector = Vector.rand
randn_vector = Vector.randn



from .misc import TactilePushingDataset, visualize_tactile_push2d
from .models import (
    TactileMeasModel,
    TactileWeightModel,
    create_tactile_models,
    get_tactile_cost_weight_inputs,
    get_tactile_initial_optim_vars,
    get_tactile_motion_capture_inputs,
    get_tactile_nn_measurements_inputs,
    get_tactile_poses_from_values,
    init_tactile_model_from_file,
    update_tactile_pushing_inputs,
)
from .pose_estimator import TactilePoseEstimator
from .trainer import TactilePushingTrainer



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Dict, Optional, Type

import torch

from theseus.core import Objective
from theseus.optimizer import Linearization
from theseus.optimizer.linear import DenseSolver, LinearSolver, LUCudaSparseSolver

from .nonlinear_least_squares import NonlinearLeastSquares

_LM_ALLOWED_ELLIPS_DAMP_SOLVERS = [DenseSolver, LUCudaSparseSolver]


def _check_ellipsoidal_damping_cls(linear_solver: LinearSolver):
    good = False
    for lsc in _LM_ALLOWED_ELLIPS_DAMP_SOLVERS:
        if isinstance(linear_solver, lsc):
            good = True
    return good


# See Nocedal and Wright, Numerical Optimization, pp. 258 - 261
# https://www.csie.ntu.edu.tw/~r97002/temp/num_optimization.pdf
class LevenbergMarquardt(NonlinearLeastSquares):
    def __init__(
        self,
        objective: Objective,
        linear_solver_cls: Optional[Type[LinearSolver]] = None,
        vectorize: bool = True,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        linear_solver_kwargs: Optional[Dict[str, Any]] = None,
        abs_err_tolerance: float = 1e-10,
        rel_err_tolerance: float = 1e-8,
        max_iterations: int = 20,
        step_size: float = 1.0,
        **kwargs,
    ):
        super().__init__(
            objective,
            linear_solver_cls=linear_solver_cls,
            vectorize=vectorize,
            linearization_cls=linearization_cls,
            linearization_kwargs=linearization_kwargs,
            linear_solver_kwargs=linear_solver_kwargs,
            abs_err_tolerance=abs_err_tolerance,
            rel_err_tolerance=rel_err_tolerance,
            max_iterations=max_iterations,
            step_size=step_size,
            **kwargs,
        )
        self._allows_ellipsoidal = _check_ellipsoidal_damping_cls(self.linear_solver)

    def compute_delta(
        self,
        damping: float = 1e-3,
        ellipsoidal_damping: bool = False,
        damping_eps: Optional[float] = None,
        **kwargs,
    ) -> torch.Tensor:

        solvers_str = ",".join(c.__name__ for c in _LM_ALLOWED_ELLIPS_DAMP_SOLVERS)
        if ellipsoidal_damping and not self._allows_ellipsoidal:
            raise NotImplementedError(
                f"Ellipsoidal damping is only supported by solvers with type "
                f"[{solvers_str}]."
            )
        if damping_eps and not self._allows_ellipsoidal:
            raise NotImplementedError(
                f"damping eps is only supported by solvers with type "
                f"[{solvers_str}]."
            )
        damping_eps = damping_eps or 1e-8

        return self.linear_solver.solve(
            damping=damping,
            ellipsoidal_damping=ellipsoidal_damping,
            damping_eps=damping_eps,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Any, List, Optional, Tuple, cast

import torch

from theseus.geometry.manifold import Manifold


# Abstract class to represent Lie groups.
# Concrete classes must implement the following methods:
#   - `exp_map`
#   - `_log_map`
#   - `_adjoint`
#   _ `_compose`
#   _ `_inverse`
#
# Constructor can optionally provide an initial tensor value as a keyword argument.
class LieGroup(Manifold):
    def __init__(
        self,
        *args: Any,
        tensor: Optional[torch.Tensor] = None,
        name: Optional[str] = None,
        dtype: torch.dtype = torch.float,
        strict: bool = False,
    ):
        super().__init__(*args, tensor=tensor, name=name, dtype=dtype, strict=strict)

    @staticmethod
    def _check_jacobians_list(jacobians: List[torch.Tensor]):
        if len(jacobians) != 0:
            raise ValueError("jacobians list to be populated must be empty.")

    @staticmethod
    @abc.abstractmethod
    def _init_tensor(*args: Any) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def dof(self) -> int:
        pass

    @staticmethod
    @abc.abstractmethod
    def rand(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "LieGroup":
        pass

    @staticmethod
    @abc.abstractmethod
    def randn(
        *size: int,
        generator: Optional[torch.Generator] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None,
        requires_grad: bool = False,
    ) -> "LieGroup":
        pass

    def __str__(self) -> str:
        return repr(self)

    @staticmethod
    @abc.abstractmethod
    def exp_map(
        tangent_vector: torch.Tensor, jacobians: Optional[List[torch.Tensor]] = None
    ) -> "LieGroup":
        pass

    @abc.abstractmethod
    def _log_map_impl(
        self, jacobians: Optional[List[torch.Tensor]] = None
    ) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def to_matrix(self) -> torch.Tensor:
        pass

    def log_map(self, jacobians: Optional[List[torch.Tensor]] = None) -> torch.Tensor:
        return self._log_map_impl(jacobians)

    @abc.abstractmethod
    def _adjoint_impl(self) -> torch.Tensor:
        pass

    def adjoint(self) -> torch.Tensor:
        return self._adjoint_impl()

    def _project_check(self, euclidean_grad: torch.Tensor, is_sparse: bool = False):
        if euclidean_grad.dtype != self.dtype:
            raise ValueError(
                "Euclidean gradients must be of the same type as the Lie group."
            )

        if euclidean_grad.device != self.device:
            raise ValueError(
                "Euclidean gradients must be on the same device as the Lie group."
            )

        if euclidean_grad.shape[-self.ndim + is_sparse :] != self.shape[is_sparse:]:
            raise ValueError(
                "Euclidean gradients must have consistent shapes with the Lie group."
            )

    def between(
        self, variable2: "LieGroup", jacobians: Optional[List[torch.Tensor]] = None
    ) -> "LieGroup":
        v1_inverse = self._inverse_impl()
        between = v1_inverse._compose_impl(variable2)
        if jacobians is not None:
            LieGroup._check_jacobians_list(jacobians)
            Jinv = LieGroup._inverse_jacobian(self)
            Jcmp0, Jcmp1 = v1_inverse._compose_jacobian(variable2)
            Jbetween0 = torch.matmul(Jcmp0, Jinv)
            jacobians.extend([Jbetween0, Jcmp1])
        return between

    @abc.abstractmethod
    def _compose_impl(self, variable2: "LieGroup") -> "LieGroup":
        pass

    def compose(
        self, variable2: "LieGroup", jacobians: Optional[List[torch.Tensor]] = None
    ) -> "LieGroup":
        composition = self._compose_impl(variable2)
        if jacobians is not None:
            LieGroup._check_jacobians_list(jacobians)
            jacobians.extend(self._compose_jacobian(variable2))
        return composition

    @abc.abstractmethod
    def _inverse_impl(self) -> "LieGroup":
        pass

    def inverse(self, jacobian: Optional[List[torch.Tensor]] = None) -> "LieGroup":
        the_inverse = self._inverse_impl()
        if jacobian is not None:
            LieGroup._check_jacobians_list(jacobian)
            jacobian.append(self._inverse_jacobian(self))
        return the_inverse

    def _compose_jacobian(
        self, group2: "LieGroup"
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        if not type(self) is type(group2):
            raise ValueError("Lie groups for compose must be of the same type.")
        g2_inverse = group2._inverse_impl()
        jac1 = g2_inverse.adjoint()
        jac2 = (
            torch.eye(group2.dof(), dtype=self.dtype)
            .repeat(group2.shape[0], 1, 1)
            .to(group2.device)
        )
        return jac1, jac2

    @staticmethod
    def _inverse_jacobian(group: "LieGroup") -> torch.Tensor:
        return -group.adjoint()

    def _local_impl(
        self, variable2: Manifold, jacobians: List[torch.Tensor] = None
    ) -> torch.Tensor:
        variable2 = cast(LieGroup, variable2)
        diff = self.between(variable2)

        if jacobians is not None:
            LieGroup._check_jacobians_list(jacobians)
            dlog: List[torch.Tensor] = []
            ret = diff.log_map(dlog)
            jacobians.append(-diff.inverse().adjoint() @ dlog[0])
            jacobians.append(dlog[0])
        else:
            ret = diff.log_map()

        return ret

    def _retract_impl(self, delta: torch.Tensor) -> "LieGroup":
        return self.compose(self.exp_map(delta))

    # added to avoid casting downstream
    def copy(self, new_name: Optional[str] = None) -> "LieGroup":
        return cast(LieGroup, super().copy(new_name=new_name))


# Alias for LieGroup.adjoint()
def adjoint(variable: LieGroup) -> torch.Tensor:
    return variable.adjoint()


def between(
    variable1: LieGroup,
    variable2: LieGroup,
    jacobians: Optional[List[torch.Tensor]] = None,
) -> LieGroup:
    return variable1.between(variable2, jacobians=jacobians)


# Alias for LieGroup.compose()
def compose(
    variable1: LieGroup,
    variable2: LieGroup,
    jacobians: Optional[List[torch.Tensor]] = None,
) -> LieGroup:
    return variable1.compose(variable2, jacobians=jacobians)


# Alias for LieGroup.inverse()
def inverse(
    variable1: LieGroup, jacobian: Optional[List[torch.Tensor]] = None
) -> LieGroup:
    return variable1.inverse(jacobian=jacobian)


# Alias for LieGroup.log_map()
def log_map(
    variable: LieGroup, jacobians: Optional[List[torch.Tensor]] = None
) -> torch.Tensor:
    return variable.log_map(jacobians=jacobians)


# Alias for LieGroup.exp_map()
def exp_map(
    variable: LieGroup,
    tangent_vector: torch.Tensor,
    jacobians: Optional[List[torch.Tensor]] = None,
) -> LieGroup:
    return variable.__class__.exp_map(tangent_vector, jacobians=jacobians)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import List, Optional, Tuple

import torch

from theseus import CostFunction, CostWeight
from theseus.geometry import LieGroup


class Local(CostFunction):
    def __init__(
        self,
        var: LieGroup,
        target: LieGroup,
        cost_weight: CostWeight,
        name: Optional[str] = None,
    ):
        super().__init__(cost_weight, name=name)
        if not isinstance(var, target.__class__):
            raise ValueError(
                "Variable for the Local inconsistent with the given target."
            )
        if not var.dof() == target.dof():
            raise ValueError(
                "Variable and target in the Local must have identical dof."
            )
        self.var = var
        self.target = target
        self.register_optim_vars(["var"])
        self.register_aux_vars(["target"])

    def error(self) -> torch.Tensor:
        return self.target.local(self.var)

    def jacobians(self) -> Tuple[List[torch.Tensor], torch.Tensor]:
        Jlist: List[torch.Tensor] = []
        self.target.local(self.var, jacobians=Jlist)
        return [Jlist[1]], self.error()

    def dim(self) -> int:
        return self.var.dof()

    def _copy_impl(self, new_name: Optional[str] = None) -> "Local":
        return Local(  # type: ignore
            self.var.copy(), self.target.copy(), self.weight.copy(), name=new_name
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import warnings
from typing import List, Optional, Sequence, Tuple, Union, cast

import torch

from .theseus_function import TheseusFunction
from .variable import Variable


# Abstract class for representing cost weights (aka, precisions, inverse covariance)
# Concrete classes must implement two methods:
#   - `weight_error`: return an error tensor weighted by the cost weight
#   - `weightJacobiansError`: returns jacobians an errors weighted by the cost weight
class CostWeight(TheseusFunction, abc.ABC):
    def __init__(
        self,
        name: Optional[str] = None,
    ):
        super().__init__(name=name)

    @abc.abstractmethod
    def weight_error(self, error: torch.Tensor) -> torch.Tensor:
        pass

    @abc.abstractmethod
    def weight_jacobians_and_error(
        self,
        jacobians: List[torch.Tensor],
        error: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        pass

    # Must copy everything
    @abc.abstractmethod
    def _copy_impl(self, new_name: Optional[str] = None) -> "TheseusFunction":
        pass

    def copy(
        self, new_name: Optional[str] = None, keep_variable_names: bool = False
    ) -> "CostWeight":
        return cast(
            CostWeight,
            super().copy(new_name=new_name, keep_variable_names=keep_variable_names),
        )


# Besides passing a theseus Variable, can also get a float and it will create the
# Variable with a default name for it
class ScaleCostWeight(CostWeight):
    def __init__(
        self,
        scale: Union[float, torch.Tensor, Variable],
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        if not isinstance(scale, Variable):
            if not isinstance(scale, torch.Tensor):
                scale = torch.tensor(scale)
            self.scale = Variable(scale)
        else:
            self.scale = scale
        if not self.scale.tensor.squeeze().ndim in [0, 1]:
            raise ValueError(
                "ScaleCostWeight only accepts 0- or 1-dim (batched) tensors."
            )
        self.scale.tensor = self.scale.tensor.view(-1, 1)
        self.register_aux_vars(["scale"])

    def weight_error(self, error: torch.Tensor) -> torch.Tensor:
        return error * self.scale.tensor

    def weight_jacobians_and_error(
        self,
        jacobians: List[torch.Tensor],
        error: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        error = error * self.scale.tensor
        new_jacobians = []
        for jac in jacobians:
            new_jacobians.append(jac * self.scale.tensor.view(-1, 1, 1))
        return new_jacobians, error

    def _copy_impl(self, new_name: Optional[str] = None) -> "ScaleCostWeight":
        return ScaleCostWeight(self.scale.copy(), name=new_name)


# Besides passing a theseus Variable, can also get any float sequence and it will create the
# Variable with a default name for it
class DiagonalCostWeight(CostWeight):
    def __init__(
        self,
        diagonal: Union[Sequence[float], torch.Tensor, Variable],
        name: Optional[str] = None,
    ):
        super().__init__(name=name)
        if not isinstance(diagonal, Variable):
            if not isinstance(diagonal, torch.Tensor):
                diagonal = torch.tensor(diagonal)
            self.diagonal = Variable(diagonal)
        else:
            self.diagonal = diagonal
        if not self.diagonal.tensor.squeeze().ndim < 3:
            raise ValueError("DiagonalCostWeight only accepts tensors with ndim < 3.")
        if self.diagonal.tensor.ndim == 0:
            self.diagonal.tensor = self.diagonal.tensor.view(1, 1)
        if self.diagonal.tensor.ndim == 1:
            warnings.warn(
                "1-D diagonal input is ambiguous. Dimension will be "
                "interpreted as dof dimension and not batch dimension."
            )
            self.diagonal.tensor = self.diagonal.tensor.view(1, -1)
        self.register_aux_vars(["diagonal"])

    def weight_error(self, error: torch.Tensor) -> torch.Tensor:
        return error * self.diagonal.tensor

    def weight_jacobians_and_error(
        self,
        jacobians: List[torch.Tensor],
        error: torch.Tensor,
    ) -> Tuple[List[torch.Tensor], torch.Tensor]:
        error = error * self.diagonal.tensor
        new_jacobians = []
        for jac in jacobians:
            # Jacobian is batch_size x cost_fuction_dim x var_dim
            # This left multiplies the weights to jacobian
            new_jacobians.append(jac * self.diagonal.tensor.unsqueeze(2))
        return new_jacobians, error

    def _copy_impl(self, new_name: Optional[str] = None) -> "DiagonalCostWeight":
        return DiagonalCostWeight(self.diagonal.copy(), name=new_name)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import numpy as np
import pytest  # noqa: F401
import torch

import theseus as th

from .common import MockCostFunction, MockCostWeight


def test_theseus_function_init():
    all_ids = []
    variables = [th.Variable(torch.ones(1, 1), name="var_1")]
    aux_vars = [th.Variable(torch.ones(1, 1), name="aux_1")]
    for i in range(100):
        cost_weight = MockCostWeight(torch.ones(1, 1), name=f"cost_weight_{i}")
        if np.random.random() < 0.5:
            name = f"name_{i}"
        else:
            name = None
        cost_function = MockCostFunction(variables, aux_vars, cost_weight, name=name)
        all_ids.append(cost_function._id)
        all_ids.append(cost_weight._id)
        if name is not None:
            assert name == cost_function.name

    assert len(set(all_ids)) == len(all_ids)


def test_no_copy_vars_check():
    variables = [th.Variable(torch.ones(1, 1), name="var_1")]
    aux_vars = [th.Variable(torch.ones(1, 1), name="aux_1")]
    cost_weight = MockCostWeight(torch.ones(1, 1), name="cost_weight")
    cost_function = MockCostFunction(
        variables, aux_vars, cost_weight, no_copy_vars=True
    )
    with pytest.raises(RuntimeError):
        cost_function.copy()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
from typing import Any, Dict, Optional, Type

import torch

from theseus.core import Objective
from theseus.optimizer import Linearization
from theseus.optimizer.linear import LinearSolver
from theseus.optimizer.linear.dense_solver import CholeskyDenseSolver

from .nonlinear_optimizer import NonlinearOptimizer


class NonlinearLeastSquares(NonlinearOptimizer, abc.ABC):
    def __init__(
        self,
        objective: Objective,
        *args,
        linear_solver_cls: Optional[Type[LinearSolver]] = None,
        vectorize: bool = True,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        linear_solver_kwargs: Optional[Dict[str, Any]] = None,
        abs_err_tolerance: float = 1e-10,
        rel_err_tolerance: float = 1e-8,
        max_iterations: int = 20,
        step_size: float = 1.0,
        **kwargs,
    ):
        linear_solver_cls = linear_solver_cls or CholeskyDenseSolver
        super().__init__(
            objective,
            linear_solver_cls=linear_solver_cls,
            vectorize=vectorize,
            linearization_cls=linearization_cls,
            linearization_kwargs=linearization_kwargs,
            linear_solver_kwargs=linear_solver_kwargs,
            abs_err_tolerance=abs_err_tolerance,
            rel_err_tolerance=rel_err_tolerance,
            max_iterations=max_iterations,
            step_size=step_size,
            **kwargs,
        )

    @abc.abstractmethod
    def compute_delta(self, **kwargs) -> torch.Tensor:
        pass



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional

import torch

from theseus.core import Objective

from .linearization import Linearization
from .variable_ordering import VariableOrdering


class DenseLinearization(Linearization):
    def __init__(
        self,
        objective: Objective,
        ordering: Optional[VariableOrdering] = None,
        **kwargs
    ):
        super().__init__(objective, ordering)
        self.A: torch.Tensor = None
        self.AtA: torch.Tensor = None
        self.b: torch.Tensor = None
        self.Atb: torch.Tensor = None

    def _linearize_jacobian_impl(self):
        err_row_idx = 0
        self.A = torch.zeros(
            (self.objective.batch_size, self.num_rows, self.num_cols),
            device=self.objective.device,
            dtype=self.objective.dtype,
        )
        self.b = torch.zeros(
            (self.objective.batch_size, self.num_rows),
            device=self.objective.device,
            dtype=self.objective.dtype,
        )
        for cost_function in self.objective._get_iterator():
            jacobians, error = cost_function.weighted_jacobians_error()
            num_rows = cost_function.dim()
            for var_idx_in_cost_function, var_jacobian in enumerate(jacobians):
                var_idx_in_order = self.ordering.index_of(
                    cost_function.optim_var_at(var_idx_in_cost_function).name
                )
                var_start_col = self.var_start_cols[var_idx_in_order]

                num_cols = var_jacobian.shape[2]
                row_slice = slice(err_row_idx, err_row_idx + num_rows)
                col_slice = slice(var_start_col, var_start_col + num_cols)
                self.A[:, row_slice, col_slice] = var_jacobian

            self.b[:, row_slice] = -error
            err_row_idx += cost_function.dim()

    def _linearize_hessian_impl(self):
        self._linearize_jacobian_impl()
        At = self.A.transpose(1, 2)
        self.AtA = At.bmm(self.A)
        self.Atb = At.bmm(self.b.unsqueeze(2))

    def hessian_approx(self):
        return self.AtA



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import abc
import warnings
from typing import Any, Dict, Optional, Type

import torch
import torch.linalg

from theseus.core import Objective
from theseus.optimizer import DenseLinearization, Linearization

from .linear_solver import LinearSolver


class DenseSolver(LinearSolver):
    def __init__(
        self,
        objective: Objective,
        linearization_cls: Optional[Type[Linearization]] = None,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        check_singular: bool = False,
    ):
        linearization_cls = linearization_cls or DenseLinearization
        if linearization_cls != DenseLinearization:
            raise RuntimeError(
                "DenseSolver only works with theseus.nonlinear.DenseLinearization, "
                f"but {linearization_cls} was provided."
            )
        super().__init__(objective, linearization_cls, linearization_kwargs)
        self.linearization: DenseLinearization = self.linearization
        self._check_singular = check_singular

    @staticmethod
    def _apply_damping(
        matrix: torch.Tensor,
        damping: float,
        ellipsoidal: bool = True,
        eps: float = 1e-8,
    ) -> torch.Tensor:
        if matrix.ndim != 3:
            raise ValueError(
                "Matrix must have a 3 dimensions, the first one being a batch dimension."
            )
        _, n, m = matrix.shape
        if n != m:
            raise ValueError("Matrix must be square.")

        # See Nocedal and Wright, Numerical Optimization, pp. 260 and 261
        # https://www.csie.ntu.edu.tw/~r97002/temp/num_optimization.pdf
        if ellipsoidal:
            # Add eps to guard against ill-conditioned matrix
            damped_D = torch.diag_embed(damping * matrix.diagonal(dim1=1, dim2=2) + eps)
        else:
            damped_D = damping * torch.eye(
                n, device=matrix.device, dtype=matrix.dtype
            ).unsqueeze(0)
        return matrix + damped_D

    def _apply_damping_and_solve(
        self,
        Atb: torch.Tensor,
        AtA: torch.Tensor,
        damping: Optional[float] = None,
        ellipsoidal_damping: bool = True,
        damping_eps: float = 1e-8,
    ) -> torch.Tensor:
        if damping:
            AtA = DenseSolver._apply_damping(
                AtA, damping, ellipsoidal=ellipsoidal_damping, eps=damping_eps
            )
        return self._solve_sytem(Atb, AtA)

    @abc.abstractmethod
    def _solve_sytem(self, Atb: torch.Tensor, AtA: torch.Tensor) -> torch.Tensor:
        pass

    def solve(
        self,
        damping: Optional[float] = None,
        ellipsoidal_damping: bool = True,
        damping_eps: float = 1e-8,
        **kwargs,
    ) -> torch.Tensor:
        if self._check_singular:
            AtA = self.linearization.AtA
            Atb = self.linearization.Atb
            with torch.no_grad():
                output = torch.zeros(AtA.shape[0], AtA.shape[1]).to(AtA.device)
                _, _, infos = torch.lu(AtA, get_infos=True)
                good_idx = infos.bool().logical_not()
                if not good_idx.all():
                    warnings.warn(
                        "Singular matrix found in batch, solution will be set "
                        "to all 0 for all singular matrices.",
                        RuntimeWarning,
                    )
            AtA = AtA[good_idx]
            Atb = Atb[good_idx]
            solution = self._apply_damping_and_solve(
                Atb,
                AtA,
                damping=damping,
                ellipsoidal_damping=ellipsoidal_damping,
                damping_eps=damping_eps,
            )
            output[good_idx] = solution
            return output
        else:
            return self._apply_damping_and_solve(
                self.linearization.Atb,
                self.linearization.AtA,
                damping=damping,
                ellipsoidal_damping=ellipsoidal_damping,
                damping_eps=damping_eps,
            )


class LUDenseSolver(DenseSolver):
    def __init__(
        self,
        objective: Objective,
        linearization_cls: Optional[Type[Linearization]] = DenseLinearization,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        check_singular: bool = False,
    ):
        super().__init__(
            objective,
            linearization_cls,
            linearization_kwargs,
            check_singular=check_singular,
        )

    def _solve_sytem(self, Atb: torch.Tensor, AtA: torch.Tensor) -> torch.Tensor:
        return torch.linalg.solve(AtA, Atb).squeeze(2)


class CholeskyDenseSolver(DenseSolver):
    def __init__(
        self,
        objective: Objective,
        linearization_cls: Optional[Type[Linearization]] = DenseLinearization,
        linearization_kwargs: Optional[Dict[str, Any]] = None,
        check_singular: bool = False,
    ):
        super().__init__(
            objective,
            linearization_cls,
            linearization_kwargs,
            check_singular=check_singular,
        )

    def _solve_sytem(self, Atb: torch.Tensor, AtA: torch.Tensor) -> torch.Tensor:
        lower = torch.linalg.cholesky(AtA)
        return torch.cholesky_solve(Atb, lower).squeeze(2)
